@@ -78,7 +78,7 @@
 -- If a predicate is not valid, 'prove' will return a counterexample: An
 -- assignment to inputs such that the predicate fails. The 'sat' function will
 -- return a satisfying assignment, if there is one. The 'allSat' function returns
--- all satisfying assignments, lazily.
+-- all satisfying assignments.
 --
 -- The sbv library uses third-party SMT solvers via the standard SMT-Lib interface:
 -- <http://smtlib.cs.uiowa.edu/>
@@ -98,6 +98,9 @@
 --
 --   * Z3 from Microsoft: <http://github.com/Z3Prover/z3/wiki>
 --
+-- SBV requires recent versions of these solvers; please see the file
+-- @SMTSolverVersions.md@ in the source distribution for specifics.
+--
 -- SBV also allows calling these solvers in parallel, either getting results from multiple solvers
 -- or returning the fastest one. (See 'proveWithAll', 'proveWithAny', etc.)
 --
@@ -105,7 +108,12 @@
 -- get in touch if there is a solver you'd like to see included.
 ---------------------------------------------------------------------------------
 
-{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE    FlexibleInstances   #-}
+{-# LANGUAGE    QuasiQuotes         #-}
+{-# LANGUAGE    TemplateHaskell     #-}
+{-# LANGUAGE    ScopedTypeVariables #-}
+{-# LANGUAGE    StandaloneDeriving  #-}
+{-# OPTIONS_GHC -fno-warn-orphans  #-}
 
 module Data.SBV (
   -- * Programming with symbolic values
@@ -122,112 +130,106 @@ module Data.SBV (
   -- *** Signed unbounded integers
   -- $unboundedLimitations
   , SInteger
-  -- *** IEEE-floating point numbers
+  -- *** Floating point numbers
   -- $floatingPoints
-  , SFloat, SDouble, IEEEFloating(..), IEEEFloatConvertable(..), RoundingMode(..), SRoundingMode, nan, infinity, sNaN, sInfinity
-  -- **** Rounding modes
-  , sRoundNearestTiesToEven, sRoundNearestTiesToAway, sRoundTowardPositive, sRoundTowardNegative, sRoundTowardZero, sRNE, sRNA, sRTP, sRTN, sRTZ
-  -- **** Bit-pattern conversions
-  , sFloatAsSWord32, sWord32AsSFloat, sDoubleAsSWord64, sWord64AsSDouble, blastSFloat, blastSDouble
+  , SFloat, SDouble
   -- *** Signed algebraic reals
   -- $algReals
   , SReal, AlgReal, sRealToSInteger
+
   -- ** Creating a symbolic variable
   -- $createSym
   , sBool, sWord8, sWord16, sWord32, sWord64, sInt8, sInt16, sInt32, sInt64, sInteger, sReal, sFloat, sDouble
+
   -- ** Creating a list of symbolic variables
   -- $createSyms
   , sBools, sWord8s, sWord16s, sWord32s, sWord64s, sInt8s, sInt16s, sInt32s, sInt64s, sIntegers, sReals, sFloats, sDoubles
+
   -- *** Abstract SBV type
-  , SBV
+  , SBV, HasKind(..), Kind(..)
   -- *** Arrays of symbolic values
   , SymArray(..), SArray, SFunArray, mkSFunArray
-  -- *** Full binary trees
-  , STree, readSTree, writeSTree, mkSTree
+
   -- ** Operations on symbolic values
   -- *** Word level
-  , sTestBit, sExtractBits, sPopCount, sShiftLeft, sShiftRight, sRotateLeft, sRotateRight, sSignedShiftArithRight, sFromIntegral, setBitTo, oneIf
-  , lsb, msb, label
-  -- *** Predicates
-  , allEqual, allDifferent, inRange, sElem
-  -- *** Addition and Multiplication with high-bits
-  , fullAdder, fullMultiplier
+  , sShiftLeft, sShiftRight, sRotateLeft, sRotateRight, sSignedShiftArithRight, sFromIntegral, oneIf
+  , label
+
   -- *** Exponentiation
   , (.^)
-  -- *** Blasting/Unblasting
-  , blastBE, blastLE, FromBits(..)
   -- *** Splitting, joining, and extending
   , Splittable(..)
-  -- ** Polynomial arithmetic and CRCs
-  , Polynomial(..), crcBV, crc
+
   -- ** Conditionals: Mergeable values
   , Mergeable(..), ite, iteLazy
-  -- ** Symbolic equality
-  , EqSymbolic(..)
-  -- ** Symbolic ordering
-  , OrdSymbolic(..)
+
   -- ** Symbolic integral numbers
   , SIntegral
+  -- ** Symbolic finite bits
+  , SFiniteBits(..)
   -- ** Division
   , SDivisible(..)
   -- ** The Boolean class
   , Boolean(..)
   -- *** Generalizations of boolean operations
   , bAnd, bOr, bAny, bAll
-  -- ** Pretty-printing and reading numbers in Hex & Binary
-  , PrettyNum(..), readBin
-  -- * Checking satisfiability in path conditions
-  , isSatisfiableInCurrentPath
 
   -- * Uninterpreted sorts, constants, and functions
   -- $uninterpreted
   , Uninterpreted(..), addAxiom
 
+  -- * Symbolic Equality and Comparisons
+  , EqSymbolic(..), OrdSymbolic(..), Equality(..)
+
+  -- * Constraints
+  -- $constrainIntro
+  , constrain, namedConstraint
+  -- ** Cardinality constraints
+  -- $cardIntro
+  , pbAtMost, pbAtLeast, pbExactly, pbLe, pbGe, pbEq, pbMutexed, pbStronglyMutexed
+
   -- * Enumerations
   -- $enumerations
+  , mkSymbolicEnumeration
 
   -- * Properties, proofs, satisfiability, and safety
   -- $proveIntro
-
-  -- ** Predicates
-  , Predicate, Provable(..), Equality(..)
-  -- ** Proving properties
-  , prove, proveWith, isTheorem, isTheoremWith
-  -- ** Checking satisfiability
-  , sat, satWith, isSatisfiable, isSatisfiableWith
+  -- $noteOnNestedQuantifiers
+  -- $multiIntro
+  , Predicate, Goal, Provable(..)
   -- ** Checking safety
   -- $safeIntro
-  , sAssert, safe, safeWith, isSafe, SExecutable(..)
-  -- ** Finding all satisfying assignments
-  , allSat, allSatWith
+  , sAssert, isSafe, SExecutable(..)
   -- ** Satisfying a sequence of boolean conditions
   , solve
-  -- ** Adding constraints
-  -- $constrainIntro
-  , constrain, pConstrain
-  -- ** Checking constraint vacuity
-  , isVacuous, isVacuousWith
+
   -- ** Quick-checking
   , sbvQuickCheck
 
-  -- * Proving properties using multiple solvers
-  -- $multiIntro
-  , proveWithAll, proveWithAny, satWithAll, satWithAny
+  -- * Running a symbolic computation
+  , runSMT, runSMTWith
+
+  -- * Solver exceptions
+  , SMTException(..)
 
   -- * Optimization
-  -- $optimizeIntro
-  , minimize, maximize, optimize
-  , minimizeWith, maximizeWith, optimizeWith
-
-  -- * Computing expected values
-  , expectedValue, expectedValueWith
+  -- $optiIntro
+  , OptimizeStyle(..), Penalty(..), Objective(..), minimize, maximize, assertSoft
+  , ExtCW(..), GeneralizedCW(..)
 
   -- * Model extraction
   -- $modelExtraction
 
   -- ** Inspecting proof results
   -- $resultTypes
-  , ThmResult(..), SatResult(..), SafeResult(..), AllSatResult(..), SMTResult(..)
+  , ThmResult(..), SatResult(..), AllSatResult(..), SafeResult(..), OptimizeResult(..), SMTResult(..)
+
+  -- * IEEE-floating point numbers
+  , IEEEFloating(..), IEEEFloatConvertable(..), RoundingMode(..), SRoundingMode, nan, infinity, sNaN, sInfinity
+  -- ** Rounding modes
+  , sRoundNearestTiesToEven, sRoundNearestTiesToAway, sRoundTowardPositive, sRoundTowardNegative, sRoundTowardZero, sRNE, sRNA, sRTP, sRTN, sRTZ
+  -- ** Bit-pattern conversions
+  , sFloatAsSWord32, sWord32AsSFloat, sDoubleAsSWord64, sWord64AsSDouble, blastSFloat, blastSDouble
 
   -- ** Programmable model extraction
   -- $programmableExtraction
@@ -235,43 +237,13 @@ module Data.SBV (
   , getModelDictionaries, getModelValues, getModelUninterpretedValues
 
   -- * SMT Interface: Configurations and solvers
-  , SMTConfig(..), SMTLibVersion(..), SMTLibLogic(..), Logic(..), OptimizeOpts(..), Solver(..), SMTSolver(..), boolector, cvc4, yices, z3, mathSAT, abc, defaultSolverConfig, sbvCurrentSolver, defaultSMTCfg, sbvCheckSolverInstallation, sbvAvailableSolvers
+  , SMTConfig(..), Timing(..), SMTLibVersion(..), Solver(..), SMTSolver(..)
+  , boolector, cvc4, yices, z3, mathSAT, abc, defaultSolverConfig, defaultSMTCfg, sbvCheckSolverInstallation, sbvAvailableSolvers
+  , setLogic, setOption, setInfo, setTimeOut
 
   -- * Symbolic computations
   , Symbolic, output, SymWord(..)
 
-  -- * Getting SMT-Lib output (for offline analysis)
-  , compileToSMTLib, generateSMTBenchmarks
-
-  -- * Test case generation
-  , genTest, getTestValues, TestVectors, TestStyle(..), renderTest, CW(..), HasKind(..), Kind(..), cwToBool
-
-  -- * Code generation from symbolic programs
-  -- $cCodeGeneration
-  , SBVCodeGen
-
-  -- ** Setting code-generation options
-  , cgPerformRTCs, cgSetDriverValues, cgGenerateDriver, cgGenerateMakefile
-
-  -- ** Designating inputs
-  , cgInput, cgInputArr
-
-  -- ** Designating outputs
-  , cgOutput, cgOutputArr
-
-  -- ** Designating return values
-  , cgReturn, cgReturnArr
-
-  -- ** Code generation with uninterpreted functions
-  , cgAddPrototype, cgAddDecl, cgAddLDFlags, cgIgnoreSAssert
-
-  -- ** Code generation with 'SInteger' and 'SReal' types
-  -- $unboundedCGen
-  , cgIntegerSize, cgSRealType, CgSRealType(..)
-
-  -- ** Compilation to C
-  , compileToC, compileToCLib
-
   -- * Module exports
   -- $moduleExportIntro
 
@@ -281,36 +253,31 @@ module Data.SBV (
   , module Data.Ratio
   ) where
 
-import Control.Monad            (filterM)
-import Control.Concurrent.Async (async, waitAny, waitAnyCancel)
-import System.IO.Unsafe         (unsafeInterleaveIO)             -- only used safely!
+import Control.Monad (filterM)
+
+import qualified Control.Exception as C
+
+import Data.SBV.Core.AlgReals
+import Data.SBV.Core.Data
+import Data.SBV.Core.Model
+import Data.SBV.Core.Floating
+import Data.SBV.Core.Splittable
 
-import Data.SBV.BitVectors.AlgReals
-import Data.SBV.BitVectors.Data
-import Data.SBV.BitVectors.Model
-import Data.SBV.BitVectors.Floating
-import Data.SBV.BitVectors.PrettyNum
-import Data.SBV.BitVectors.Splittable
-import Data.SBV.BitVectors.STree
-import Data.SBV.Compilers.C
-import Data.SBV.Compilers.CodeGen
 import Data.SBV.Provers.Prover
-import Data.SBV.Tools.GenTest
-import Data.SBV.Tools.ExpectedValue
-import Data.SBV.Tools.Optimize
-import Data.SBV.Tools.Polynomial
+
 import Data.SBV.Utils.Boolean
+import Data.SBV.Utils.TDiff   (Timing(..))
+
 import Data.Bits
 import Data.Int
 import Data.Ratio
 import Data.Word
 
--- | The currently active solver, obtained by importing "Data.SBV".
--- To have other solvers /current/, import one of the bridge
--- modules "Data.SBV.Bridge.ABC", "Data.SBV.Bridge.Boolector", "Data.SBV.Bridge.CVC4",
--- "Data.SBV.Bridge.Yices", or "Data.SBV.Bridge.Z3" directly.
-sbvCurrentSolver :: SMTConfig
-sbvCurrentSolver = z3
+import qualified Language.Haskell.TH as TH
+import Data.Generics
+
+import Data.SBV.SMT.Utils (SMTException(..))
+import Data.SBV.Control.Utils (SMTValue)
 
 -- | Form the symbolic conjunction of a given list of boolean conditions. Useful in expressing
 -- problems with constraints, like the following:
@@ -325,10 +292,11 @@ solve = return . bAnd
 -- | Check whether the given solver is installed and is ready to go. This call does a
 -- simple call to the solver to ensure all is well.
 sbvCheckSolverInstallation :: SMTConfig -> IO Bool
-sbvCheckSolverInstallation cfg = do ThmResult r <- proveWith cfg $ \x -> (x+x) .== ((x*2) :: SWord8)
-                                    case r of
-                                      Unsatisfiable _ -> return True
-                                      _               -> return False
+sbvCheckSolverInstallation cfg = check `C.catch` (\(_ :: C.SomeException) -> return False)
+  where check = do ThmResult r <- proveWith cfg $ \x -> (x+x) .== ((x*2) :: SWord8)
+                   case r of
+                     Unsatisfiable{} -> return True
+                     _               -> return False
 
 -- | The default configs corresponding to supported SMT solvers
 defaultSolverConfig :: Solver -> SMTConfig
@@ -343,43 +311,13 @@ defaultSolverConfig ABC       = abc
 sbvAvailableSolvers :: IO [SMTConfig]
 sbvAvailableSolvers = filterM sbvCheckSolverInstallation (map defaultSolverConfig [minBound .. maxBound])
 
-sbvWithAny :: [SMTConfig] -> (SMTConfig -> a -> IO b) -> a -> IO (Solver, b)
-sbvWithAny []      _    _ = error "SBV.withAny: No solvers given!"
-sbvWithAny solvers what a = snd `fmap` (mapM try solvers >>= waitAnyCancel)
-   where try s = async $ what s a >>= \r -> return (name (solver s), r)
-
-sbvWithAll :: [SMTConfig] -> (SMTConfig -> a -> IO b) -> a -> IO [(Solver, b)]
-sbvWithAll solvers what a = mapM try solvers >>= (unsafeInterleaveIO . go)
-   where try s = async $ what s a >>= \r -> return (name (solver s), r)
-         go []  = return []
-         go as  = do (d, r) <- waitAny as
-                     -- The following filter works because the Eq instance on Async
-                     -- checks the thread-id; so we know that we're removing the
-                     -- correct solver from the list. This also allows for
-                     -- running the same-solver (with different options), since
-                     -- they will get different thread-ids.
-                     rs <- unsafeInterleaveIO $ go (filter (/= d) as)
-                     return (r : rs)
-
--- | Prove a property with multiple solvers, running them in separate threads. The
--- results will be returned in the order produced.
-proveWithAll :: Provable a => [SMTConfig] -> a -> IO [(Solver, ThmResult)]
-proveWithAll  = (`sbvWithAll` proveWith)
-
--- | Prove a property with multiple solvers, running them in separate threads. Only
--- the result of the first one to finish will be returned, remaining threads will be killed.
-proveWithAny :: Provable a => [SMTConfig] -> a -> IO (Solver, ThmResult)
-proveWithAny  = (`sbvWithAny` proveWith)
-
--- | Find a satisfying assignment to a property with multiple solvers, running them in separate threads. The
--- results will be returned in the order produced.
-satWithAll :: Provable a => [SMTConfig] -> a -> IO [(Solver, SatResult)]
-satWithAll = (`sbvWithAll` satWith)
-
--- | Find a satisfying assignment to a property with multiple solvers, running them in separate threads. Only
--- the result of the first one to finish will be returned, remaining threads will be killed.
-satWithAny :: Provable a => [SMTConfig] -> a -> IO (Solver, SatResult)
-satWithAny    = (`sbvWithAny` satWith)
+-- If we get a program producing nothing (i.e., Symbolic ()), pretend it simply returns True.
+-- This is useful since min/max calls and constraints will provide the context
+instance Provable Goal where
+  forAll_    a = forAll_    ((a >> return true) :: Predicate)
+  forAll ns  a = forAll ns  ((a >> return true) :: Predicate)
+  forSome_   a = forSome_   ((a >> return true) :: Predicate)
+  forSome ns a = forSome ns ((a >> return true) :: Predicate)
 
 -- | Equality as a proof method. Allows for
 -- very concise construction of equivalence proofs, which is very typical in
@@ -429,6 +367,21 @@ instance {-# OVERLAPPABLE #-}
 instance {-# OVERLAPPABLE #-} (SymWord a, SymWord b, SymWord c, SymWord d, SymWord e, SymWord f, SymWord g, EqSymbolic z) => Equality ((SBV a, SBV b, SBV c, SBV d, SBV e, SBV f, SBV g) -> z) where
   k === l = prove $ \a b c d e f g -> k (a, b, c, d, e, f, g) .== l (a, b, c, d, e, f, g)
 
+-- | Make an enumeration a symbolic type.
+mkSymbolicEnumeration :: TH.Name -> TH.Q [TH.Dec]
+mkSymbolicEnumeration typeName = do
+    let typeCon = TH.conT typeName
+    [d| deriving instance Eq       $(typeCon)
+        deriving instance Show     $(typeCon)
+        deriving instance Ord      $(typeCon)
+        deriving instance Read     $(typeCon)
+        deriving instance Data     $(typeCon)
+        deriving instance SymWord  $(typeCon)
+        deriving instance HasKind  $(typeCon)
+        deriving instance SMTValue $(typeCon)
+        deriving instance SatModel $(typeCon)
+      |]
+
 -- Haddock section documentation
 {- $progIntro
 The SBV library is really two things:
@@ -450,6 +403,15 @@ particular, any Haskell program you build out of symbolic components is fully co
 executable within Haskell, without the need for any custom interpreters. (They are truly
 Haskell programs, not AST's built out of pieces of syntax.) This provides for an integrated
 feel of the system, one of the original design goals for SBV.
+
+= Incremental mode: Queries
+
+SBV provides a wide variety of ways to utilize SMT-solvers, without requiring the user to
+deal with the solvers themselves. While this mode is convenient, advanced users might need
+access to the underlying solver, using the SMTLib language. For such use cases, SBV allows
+users to have an interactive session: The user can issue commands to the solver, inspect
+the values/results, and formulate new constraints. This advanced feature is available through
+the "Data.SBV.Control" module, where most SMTLib features are made available via a typed-API.
 -}
 
 {- $proveIntro
@@ -503,8 +465,8 @@ We can use 'safe' to statically see if such a violation is possible before we us
 
 >>> safe (sub :: SInt8 -> SInt8 -> SInt8)
 [sub: x >= y must hold!: Violated. Model:
-  s0 = -128 :: Int8
-  s1 = -127 :: Int8]
+  s0 = 30 :: Int8
+  s1 = 32 :: Int8]
 
 What happens if we make sure to arrange for this invariant? Consider this version:
 
@@ -525,47 +487,114 @@ Also see "Data.SBV.Examples.Misc.NoDiv0" for the classic div-by-zero example.
 -}
 
 
-{- $optimizeIntro
-Symbolic optimization. A call of the form:
+{- $optiIntro
+  SBV can optimize metric functions, i.e., those that generate both bounded 'SIntN', 'SWordN', and unbounded 'SInteger'
+  types, along with those produce 'SReal's. That is, it can find models satisfying all the constraints while minimizing
+  or maximizing user given metrics. Currently, optimization requires the use of the z3 SMT solver as the backend,
+  and a good review of these features is given
+  in this paper: <http://www.easychair.org/publications/download/Z_-_Maximal_Satisfaction_with_Z3>.
 
-    @minimize Quantified cost n valid@
+  Goals can be lexicographically (default), independently, or pareto-front optimized. The relevant functions are:
 
-returns @Just xs@, such that:
+      * 'minimize': Minimize a given arithmetic goal
+      * 'maximize': Minimize a given arithmetic goal
 
-   * @xs@ has precisely @n@ elements
+  Goals can be optimized at a regular or an extended value: An extended value is either positive or negative infinity
+  (for unbounded integers and reals) or positive or negative epsilon differential from a real value (for reals).
 
-   * @valid xs@ holds
+  For instance, a call of the form 
 
-   * @cost xs@ is minimal. That is, for all sequences @ys@ that satisfy the first two criteria above, @cost xs .<= cost ys@ holds.
+       @ 'minimize' "name-of-goal" $ x + 2*y @
 
-If there is no such sequence, then 'minimize' will return 'Nothing'.
+  minimizes the arithmetic goal @x+2*y@, where @x@ and @y@ can be signed\/unsigned bit-vectors, reals,
+  or integers.
 
-The function 'maximize' is similar, except the comparator is '.>='. So the value returned has the largest cost (or value, in that case).
+== A simple example
 
-The function 'optimize' allows the user to give a custom comparison function.
+  Here's an optimization example in action:
 
-The 'OptimizeOpts' argument controls how the optimization is done. If 'Quantified' is used, then the SBV optimization engine satisfies the following predicate:
+  >>> optimize Lexicographic $ \x y -> minimize "goal" (x+2*(y::SInteger))
+  Optimal in an extension field:
+    goal = -oo :: Integer
 
-   @exists xs. forall ys. valid xs && (valid ys \`implies\` (cost xs \`cmp\` cost ys))@
+  We will describe the role of the constructor 'Lexicographic' shortly.
 
-Note that this may cause efficiency problems as it involves alternating quantifiers.
-If 'OptimizeOpts' is set to 'Iterative' 'True', then SBV will programmatically
-search for an optimal solution, by repeatedly calling the solver appropriately. (The boolean argument controls whether progress reports are given. Use
-'False' for quiet operation.)
+  Of course, this becomes more useful when the result is not in an extension field:
 
-=== Quantified vs Iterative
+>>> :{
+    optimize Lexicographic $ do
+                  x <- sInteger "x"
+                  y <- sInteger "y"
+                  constrain $ x .> 0
+                  constrain $ x .< 6
+                  constrain $ y .> 2
+                  constrain $ y .< 12
+                  minimize "goal" $ x + 2 * y
+    :}
+Optimal model:
+  x    = 1 :: Integer
+  y    = 3 :: Integer
+  goal = 7 :: Integer
 
-Note that the quantified and iterative versions are two different optimization approaches and may not necessarily yield the same
-results. In particular, the quantified version can tell us no such solution exists if there is no global optimum value, while the iterative
-version might simply loop forever for such a problem. To wit, consider the example:
+  As usual, the programmatic API can be used to extract the values of objectives and model-values ('getModelObjectives',
+  'getModelAssignment', etc.) to access these values and program with them further.
 
-   @ maximize Quantified head 1 (const true :: [SInteger] -> SBool) @
+== Multiple optimization goals
 
-which asks for the largest `SInteger` value. The SMT solver will happily answer back saying there is no such value with the 'Quantified' call, but the 'Iterative' variant
-will simply loop forever as it would search through an infinite chain of ascending 'SInteger' values.
+  Multiple goals can be specified, using the same syntax. In this case, the user gets to pick what style of
+  optimization to perform, by passing the relevant 'OptimizeStyle' as the first argument to 'optimize'.
 
-In practice, however, the iterative version is usually the more effective choice since alternating quantifiers are hard to deal with for many SMT-solvers and thus will
-likely result in an @unknown@ result. While the 'Iterative' variant can loop for a long time, one can simply use the boolean flag 'True' and see how the search is progressing.
+    * ['Lexicographic']. The solver will optimize the goals in the given order, optimizing
+      the latter ones under the model that optimizes the previous ones.
+
+    * ['Independent']. The solver will optimize the goals independently of each other. In this case the user will
+      be presented a model for each goal given.
+
+    * ['Pareto']. Finally, the user can query for pareto-fronts. A pareto front is an model such that no goal can be made
+      "better" without making some other goal "worse."
+
+      The optional number argument to 'Pareto' specifies the maximum number of pareto-fronts the user is asking
+      to get. If 'Nothing', SBV will query for all pareto-fronts. Note that pareto-fronts can be infinite
+      in number, so if 'Nothing' is used, there is a potential for infinitely waiting for the SBV-solver interaction
+      to finish. (If you suspect this might be the case, run in 'verbose' mode to see the interaction and
+      put a limiting factor appropriately.)
+
+== Soft Assertions
+
+  Related to optimization, SBV implements soft-asserts via 'assertSoft' calls. A soft assertion
+  is a hint to the SMT solver that we would like a particular condition to hold if **possible*.
+  That is, if there is a solution satisfying it, then we would like it to hold, but it can be violated
+  if there is no way to satisfy it. Each soft-assertion can be associated with a numeric penalty for
+  not satisfying it, hence turning it into an optimization problem.
+
+  Note that 'assertSoft' works well with optimization goals ('minimize'/'maximize' etc.),
+  and are most useful when we are optimizing a metric and thus some of the constraints
+  can be relaxed with a penalty to obtain a good solution. Again
+  see <http://www.easychair.org/publications/download/Z_-_Maximal_Satisfaction_with_Z3>
+  for a good overview of the features in Z3 that SBV is providing the bridge for.
+
+  A soft assertion can be specified in one of the following three main ways:
+
+       @
+         'assertSoft' "bounded_x" (x .< 5) 'DefaultPenalty'
+         'assertSoft' "bounded_x" (x .< 5) ('Penalty' 2.3 Nothing)
+         'assertSoft' "bounded_x" (x .< 5) ('Penalty' 4.7 (Just "group-1")) @
+
+  In the first form, we are saying that the constraint @x .< 5@ must be satisfied, if possible,
+  but if this constraint can not be satisfied to find a model, it can be violated with the default penalty of 1.
+
+  In the second case, we are associating a penalty value of @2.3@.
+
+  Finally in the third case, we are also associating this constraint with a group. The group
+  name is only needed if we have classes of soft-constraints that should be considered together.
+
+== Optimization examples
+
+  The following examples illustrate the use of basic optimization routines:
+
+     * "Data.SBV.Examples.Optimization.LinearOpt": Simple linear-optimization example.
+     * "Data.SBV.Examples.Optimization.Production": Scheduling machines in a shop
+     * "Data.SBV.Examples.Optimization.VM": Scheduling virtual-machines in a data-center
 -}
 
 {- $modelExtraction
@@ -587,26 +616,6 @@ for library construction) that the SMT-models are reinterpreted in terms of doma
 extraction allows getting arbitrarily typed models out of SMT models.
 -}
 
-{- $cCodeGeneration
-The SBV library can generate straight-line executable code in C. (While other target languages are
-certainly possible, currently only C is supported.) The generated code will perform no run-time memory-allocations,
-(no calls to @malloc@), so its memory usage can be predicted ahead of time. Also, the functions will execute precisely the
-same instructions in all calls, so they have predictable timing properties as well. The generated code
-has no loops or jumps, and is typically quite fast. While the generated code can be large due to complete unrolling,
-these characteristics make them suitable for use in hard real-time systems, as well as in traditional computing.
--}
-
-{- $unboundedCGen
-The types 'SInteger' and 'SReal' are unbounded quantities that have no direct counterparts in the C language. Therefore,
-it is not possible to generate standard C code for SBV programs using these types, unless custom libraries are available. To
-overcome this, SBV allows the user to explicitly set what the corresponding types should be for these two cases, using
-the functions below. Note that while these mappings will produce valid C code, the resulting code will be subject to
-overflow/underflows for 'SInteger', and rounding for 'SReal', so there is an implicit loss of precision.
-
-If the user does /not/ specify these mappings, then SBV will
-refuse to compile programs that involve these types.
--}
-
 {- $moduleExportIntro
 The SBV library exports the following modules wholesale, as user programs will have to import these
 modules to make any sensible use of the SBV functionality.
@@ -681,7 +690,7 @@ the introduction of variables.
 Note that the proper reading of a constraint
 depends on the context:
 
-    * In a 'sat' (or 'allSat') call: The constraint added is asserted
+  * In a 'sat' (or 'allSat') call: The constraint added is asserted
     conjunctively. That is, the resulting satisfying model (if any) will
     always satisfy all the constraints given.
 
@@ -724,31 +733,61 @@ Also note that this semantics imply that test case generation ('genTest') and qu
 can take arbitrarily long in the presence of constraints, if the random input values generated
 rarely satisfy the constraints. (As an extreme case, consider @'constrain' 'false'@.)
 
-A probabilistic constraint (see 'pConstrain') attaches a probability threshold for the
-constraint to be considered. For instance:
+=== Named constraints and unsat cores
 
-  @
-     'pConstrain' 0.8 c
-  @
+Constraints can be given names:
 
-will make sure that the condition @c@ is satisfied 80% of the time (and correspondingly, falsified 20%
-of the time), in expectation. This variant is useful for 'genTest' and 'quickCheck' functions, where we
-want to filter the test cases according to some probability distribution, to make sure that the test-vectors
-are drawn from interesting subsets of the input space. For instance, if we were to generate 100 test cases
-with the above constraint, we'd expect about 80 of them to satisfy the condition @c@, while about 20 of them
-will fail it.
+  @ 'namedConstraint' "a is at least 5" $ a .>= 5@
 
-The following properties hold:
+Such constraints are useful when used in conjunction with 'getUnsatCore' function
+where the backend solver can be queried to obtain an unsat core in case the constraints are unsatisfiable.
+This feature is enabled by the following option:
 
-  @
-    'constrain'      = 'pConstrain' 1
-    'pConstrain' t c = 'pConstrain' (1-t) (not c)
-  @
+   @ setOption $ ProduceUnsatCores True @
 
-Note that while 'constrain' can be used freely, 'pConstrain' is only allowed in the contexts of
-'genTest' or 'quickCheck'. Calls to 'pConstrain' in a prove/sat call will be rejected as SBV does not
-deal with probabilistic constraints when it comes to satisfiability and proofs.
-Also, both 'constrain' and 'pConstrain' calls during code-generation will also be rejected, for similar reasons.
+See "Data.SBV.Examples.Misc.UnsatCore" for an example use case.
+
+=== Constraint vacuity
+
+When adding constraints, one has to be careful about
+making sure they are not inconsistent. The function 'isVacuous' can be use for this purpose.
+Here is an example. Consider the following predicate:
+
+    >>> let pred = do { x <- free "x"; constrain $ x .< x; return $ x .>= (5 :: SWord8) }
+
+This predicate asserts that all 8-bit values are larger than 5, subject to the constraint that the
+values considered satisfy @x .< x@, i.e., they are less than themselves. Since there are no values that
+satisfy this constraint, the proof will pass vacuously:
+
+    >>> prove pred
+    Q.E.D.
+
+We can use 'isVacuous' to make sure to see that the pass was vacuous:
+
+    >>> isVacuous pred
+    True
+
+While the above example is trivial, things can get complicated if there are multiple constraints with
+non-straightforward relations; so if constraints are used one should make sure to check the predicate
+is not vacuously true. Here's an example that is not vacuous:
+
+     >>> let pred' = do { x <- free "x"; constrain $ x .> 6; return $ x .>= (5 :: SWord8) }
+
+This time the proof passes as expected:
+
+     >>> prove pred'
+     Q.E.D.
+
+And the proof is not vacuous:
+
+     >>> isVacuous pred'
+     False
+
+=== Checking for vacuity
+
+As we discussed SBV does not check that a given constraints is not vacuous. That is, that it can never be satisfied. This is usually
+the right behavior, since checking vacuity can be costly. The functions 'isVacuous' and 'isVacuousWith' should be used
+to explicitly check for constraint vacuity if desired.
 -}
 
 {- $uninterpreted
@@ -762,7 +801,9 @@ following example demonstrates:
 (Note that you'll also need to use the language pragmas @DeriveDataTypeable@, @DeriveAnyClass@, and import @Data.Generics@ for the above to work.) 
 
 This is all it takes to introduce 'B' as an uninterpreted sort in SBV, which makes the type @SBV B@ automagically become available as the type
-of symbolic values that ranges over 'B' values. Note that the @()@ argument is important to distinguish it from enumerations.
+of symbolic values that ranges over 'B' values. Note that the @()@ argument is important to distinguish it from enumerations, which will be
+translated to proper SMT data-types.
+
 
 Uninterpreted functions over both uninterpreted and regular sorts can be declared using the facilities introduced by
 the 'Uninterpreted' class.
@@ -772,15 +813,24 @@ the 'Uninterpreted' class.
 If the uninterpreted sort definition takes the form of an enumeration (i.e., a simple data type with all nullary constructors), then SBV will actually
 translate that as just such a data-type to SMT-Lib, and will use the constructors as the inhabitants of the said sort. A simple example is:
 
-  @
-    data X = A | B | C deriving (Eq, Ord, Show, Read, Data, SymWord, HasKind, SatModel)
-  @
+@
+    data X = A | B | C
+    mkSymbolicEnumeration ''X
+@
+
+Note the magic incantation @mkSymbolicEnumeration ''X@. For this to work, you need to have the following
+options turned on:
+
+>   LANGUAGE TemplateHaskell
+>   LANGUAGE StandaloneDeriving
+>   LANGUAGE DeriveDataTypeable
+>   LANGUAGE DeriveAnyClass
 
 Now, the user can define
 
-  @
+@
     type SX = SBV X
-  @
+@
 
 and treat @SX@ as a regular symbolic type ranging over the values @A@, @B@, and @C@. Such values can be compared for equality, and with the usual
 other comparison operators, such as @.==@, @./=@, @.>@, @.>=@, @<@, and @<=@.
@@ -788,13 +838,13 @@ other comparison operators, such as @.==@, @./=@, @.>@, @.>=@, @<@, and @<=@.
 Note that in this latter case the type is no longer uninterpreted, but is properly represented as a simple enumeration of the said elements. A simple
 query would look like:
 
-   @
+@
      allSat $ \x -> x .== (x :: SX)
-   @
+@
 
 which would list all three elements of this domain as satisfying solutions.
 
-   @
+@
      Solution #1:
        s0 = A :: X
      Solution #2:
@@ -802,10 +852,51 @@ which would list all three elements of this domain as satisfying solutions.
      Solution #3:
        s0 = C :: X
      Found 3 different solutions.
-   @
+@
 
-Note that the result is properly typed as @X@ elements; these are not mere strings. So, in a 'getModel' scenario, the user can recover actual
+Note that the result is properly typed as @X@ elements; these are not mere strings. So, in a 'getModelAssignment' scenario, the user can recover actual
 elements of the domain and program further with those values as usual.
+
+See "Data.SBV.Examples.Misc.Enumerate" for an extended example on how to use symbolic enumerations.
+-}
+
+{- $noteOnNestedQuantifiers
+=== A note on reasoning in the presence of quantifers
+
+Note that SBV allows reasoning with quantifiers: Inputs can be existentially or universally quantified. Predicates can be built
+with arbitrary nesting of such quantifiers as well. However, SBV always /assumes/ that the input is in
+prenex-normal form: <https://en.wikipedia.org/wiki/Prenex_normal_form>. That is,
+all the input declarations are treated as happening at the beginning of a predicate, followed by the actual formula. Unfortunately,
+the way predicates are written can be misleading at times, since symbolic inputs can be created at arbitrary points; interleaving them
+with other code. The rule is simple, however: All inputs are assumed at the top, in the order declared, regardless of their quantifiers.
+SBV will apply skolemization to get rid of existentials before sending predicates to backend solvers. However, if you do want nested
+quantification, you will manually have to first convert to prenex-normal form (which produces an equisatisfiable but not necessarily
+equivalent formula), and code that explicitly in SBV. See <https://github.com/LeventErkok/sbv/issues/256> for a detailed discussion
+of this issue.
+-}
+
+{- $cardIntro
+A pseudo-boolean function (<http://en.wikipedia.org/wiki/Pseudo-Boolean_function>) is a
+function from booleans to reals, basically treating 'True' as @1@ and 'False' as @0@. They
+are typically expressed in polynomial form. Such functions can be used to express cardinality
+constraints, where we want to /count/ how many things satisfy a certain condition.
+
+One can code such constraints using regular SBV programming: Simply
+walk over the booleans and the corresponding coefficients, and assert the required relation.
+For instance:
+
+   > [b0, b1, b2, b3] `pbAtMost` 2
+
+is precisely equivalent to:
+
+   > sum (map (\b -> ite b 1 0) [b0, b1, b2, b3]) .<= 2
+
+and they both express that at most /two/ of @b0@, @b1@, @b2@, and @b3@ can be 'true'.
+However, the equivalent forms give rise to long formulas and the cardinality constraint
+can get lost in the translation. The idea here is that if you use these functions instead, SBV will
+produce better translations to SMTLib for more efficient solving of cardinality constraints, assuming
+the backend solver supports them. Currently, only Z3 supports pseudo-booleans directly. For all other solvers,
+SBV will translate these to equivalent terms that do not require special functions.
 -}
 
 {-# ANN module ("HLint: ignore Use import/export shortcut" :: String) #-}
