@@ -16,7 +16,7 @@ module Data.SBV.Compilers.C(compileToC, compileToCLib, compileToC', compileToCLi
 
 import Control.DeepSeq                (rnf)
 import Data.Char                      (isSpace)
-import Data.List                      (nub, intercalate)
+import Data.List                      (nub, intercalate, intersperse)
 import Data.Maybe                     (isJust, isNothing, fromJust)
 import qualified Data.Foldable as F   (toList)
 import qualified Data.Set      as Set (member, union, unions, empty, toList, singleton, fromList)
@@ -24,14 +24,12 @@ import System.FilePath                (takeBaseName, replaceExtension)
 import System.Random
 import Text.PrettyPrint.HughesPJ
 
-import Data.SBV.BitVectors.Data
-import Data.SBV.BitVectors.PrettyNum (shex, showCFloat, showCDouble)
+import Data.SBV.Core.Data
 import Data.SBV.Compilers.CodeGen
 
-import GHC.Stack.Compat
-#if !MIN_VERSION_base(4,9,0)
-import GHC.SrcLoc.Compat
-#endif
+import Data.SBV.Utils.PrettyNum   (shex, showCFloat, showCDouble)
+
+import GHC.Stack
 
 ---------------------------------------------------------------------------
 -- * API
@@ -407,7 +405,7 @@ genDriver cfg randVals fn inps outs mbRet = [pre, header, body, post]
 
 -- | Generate the C program
 genCProg :: CgConfig -> String -> Doc -> Result -> [(String, CgVal)] -> [(String, CgVal)] -> Maybe SW -> Doc -> [Doc]
-genCProg cfg fn proto (Result kindInfo _tvals cgs ins preConsts tbls arrs _ _ (SBVPgm asgns) cstrs origAsserts _) inVars outVars mbRet extDecls
+genCProg cfg fn proto (Result kindInfo _tvals cgs ins preConsts tbls arrs _uis _axioms (SBVPgm asgns) cstrs origAsserts _) inVars outVars mbRet extDecls
   | isNothing (cgInteger cfg) && KUnbounded `Set.member` kindInfo
   = error $ "SBV->C: Unbounded integers are not supported by the C compiler."
           ++ "\nUse 'cgIntegerSize' to specify a fixed size for SInteger representation."
@@ -420,16 +418,20 @@ genCProg cfg fn proto (Result kindInfo _tvals cgs ins preConsts tbls arrs _ _ (S
   = tbd "Explicit constraints"
   | not (null arrs)
   = tbd "User specified arrays"
-  | needsExistentials (map fst ins)
+  | needsExistentials (map fst (fst ins))
   = error "SBV->C: Cannot compile functions with existentially quantified variables."
   | True
   = [pre, header, post]
  where asserts | cgIgnoreAsserts cfg = []
                | True                = origAsserts
+
        usorts = [s | KUserSort s _ <- Set.toList kindInfo, s /= "RoundingMode"] -- No support for any sorts other than RoundingMode!
+
        pre    =  text "/* File:" <+> doubleQuotes (nm <> text ".c") <> text ". Automatically generated by SBV. Do not edit! */"
               $$ text ""
+
        header = text "#include" <+> doubleQuotes (nm <> text ".h")
+
        post   = text ""
              $$ vcat (map codeSeg cgs)
              $$ extDecls
@@ -444,12 +446,16 @@ genCProg cfg fn proto (Result kindInfo _tvals cgs ins preConsts tbls arrs _ _ (S
                        )
              $$ text "}"
              $$ text ""
+
        nm = text fn
+
        assignments = F.toList asgns
+
        codeSeg (fnm, ls) =  text "/* User specified custom code for" <+> doubleQuotes (text fnm) <+> text "*/"
                          $$ vcat (map text ls)
                          $$ text ""
-       typeWidth = getMax 0 $ [len (kindOf s) | (s, _) <- assignments] ++ [len (kindOf s) | (_, (s, _)) <- ins]
+
+       typeWidth = getMax 0 $ [len (kindOf s) | (s, _) <- assignments] ++ [len (kindOf s) | (_, (s, _)) <- fst ins]
                 where len KReal{}             = 5
                       len KFloat{}            = 6 -- SFloat
                       len KDouble{}           = 7 -- SDouble
@@ -461,21 +467,28 @@ genCProg cfg fn proto (Result kindInfo _tvals cgs ins preConsts tbls arrs _ _ (S
                       getMax 8 _      = 8  -- 8 is the max we can get with SInteger, so don't bother looking any further
                       getMax m []     = m
                       getMax m (x:xs) = getMax (m `max` x) xs
+
        consts = (falseSW, falseCW) : (trueSW, trueCW) : preConsts
+
        isConst s = isJust (lookup s consts)
+
        -- TODO: The following is brittle. We should really have a function elsewhere
        -- that walks the SBVExprs and collects the SWs together.
        usedVariables = Set.unions (retSWs : map usedCgVal outVars ++ map usedAsgn assignments)
          where retSWs = maybe Set.empty Set.singleton mbRet
+
                usedCgVal (_, CgAtomic s)  = Set.singleton s
                usedCgVal (_, CgArray ss)  = Set.fromList ss
                usedAsgn  (_, SBVApp o ss) = Set.union (opSWs o) (Set.fromList ss)
+
                opSWs (LkUp _ a b)             = Set.fromList [a, b]
                opSWs (IEEEFP (FP_Cast _ _ s)) = Set.singleton s
                opSWs _                        = Set.empty
+
        isAlive :: (String, CgVal) -> Bool
        isAlive (_, CgAtomic sw) = sw `Set.member` usedVariables
        isAlive (_, _)           = True
+
        genIO :: Bool -> (Bool, (String, CgVal)) -> [Doc]
        genIO True  (alive, (cNm, CgAtomic sw)) = [declSW typeWidth sw  <+> text "=" <+> text cNm <> semi             | alive]
        genIO False (alive, (cNm, CgAtomic sw)) = [text "*" <> text cNm <+> text "=" <+> showSW cfg consts sw <> semi | alive]
@@ -484,15 +497,19 @@ genCProg cfg fn proto (Result kindInfo _tvals cgs ins preConsts tbls arrs _ _ (S
                  | isInp = declSW typeWidth sw <+> text "=" <+> text entry       <> semi
                  | True  = text entry          <+> text "=" <+> showSW cfg consts sw <> semi
                  where entry = cNm ++ "[" ++ show i ++ "]"
+
        mkRet sw = text "return" <+> showSW cfg consts sw <> semi
+
        genTbl :: ((Int, Kind, Kind), [SW]) -> (Int, Doc)
        genTbl ((i, _, k), elts) =  (location, static <+> text "const" <+> text (show k) <+> text ("table" ++ show i) <> text "[] = {"
                                               $$ nest 4 (fsep (punctuate comma (align (map (showSW cfg consts) elts))))
                                               $$ text "};")
          where static   = if location == -1 then text "static" else empty
                location = maximum (-1 : map getNodeId elts)
+
        getNodeId s@(SW _ (NodeId n)) | isConst s = -1
                                      | True      = n
+
        genAsgn :: (SW, SBVExpr) -> (Int, Doc)
        genAsgn (sw, n) = (getNodeId sw, ppExpr cfg consts n (declSW typeWidth sw) (declSWNoConst typeWidth sw) <> semi)
 
@@ -505,6 +522,7 @@ genCProg cfg fn proto (Result kindInfo _tvals cgs ins preConsts tbls arrs _ _ (S
                merge2 ts@((i, t):trest) as@((i', a):arest)
                  | i < i'                                 = (i,  t)  : merge2 trest as
                  | True                                   = (i', a) : merge2 ts arest
+
        genAssert (msg, cs, sw) = (getNodeId sw, doc)
          where doc =     text "/* ASSERTION:" <+> text msg
                      $$  maybe empty (vcat . map text) (locInfo (getCallStack `fmap` cs))
@@ -521,6 +539,18 @@ genCProg cfg fn proto (Result kindInfo _tvals cgs ins preConsts tbls arrs _ _ (S
                                          (f:rs) -> Just $ (" * SOURCE   : " ++ f) : map (" *            " ++)  rs
                locInfo _         = Nothing
 
+handlePB :: PBOp -> [Doc] -> Doc
+handlePB o args = case o of
+                    PB_AtMost  k -> addIf (repeat 1) <+> text "<=" <+> int k
+                    PB_AtLeast k -> addIf (repeat 1) <+> text ">=" <+> int k
+                    PB_Exactly k -> addIf (repeat 1) <+> text "==" <+> int k
+                    PB_Le cs   k -> addIf cs         <+> text "<=" <+> int k
+                    PB_Ge cs   k -> addIf cs         <+> text ">=" <+> int k
+                    PB_Eq cs   k -> addIf cs         <+> text "==" <+> int k
+
+  where addIf :: [Int] -> Doc
+        addIf cs = parens $ fsep $ intersperse (text "+") [parens (a <+> text "?" <+> int c <+> text ":" <+> int 0) | (a, c) <- zip args cs]
+
 handleIEEE :: FPOp -> [(SW, CW)] -> [(SW, Doc)] -> Doc -> Doc
 handleIEEE w consts as var = cvt w
   where same f                   = (f, f)
@@ -622,17 +652,28 @@ ppExpr cfg consts (SBVApp op opArgs) lhs (typ, var)
   = lhs <+> text "=" <+> rhs
   where doNotAssign (IEEEFP FP_Reinterpret{}) = True   -- generates a memcpy instead; no simple assignment
         doNotAssign _                         = False  -- generates simple assignment
+
         rhs = p op (map (showSW cfg consts) opArgs)
+
         rtc = cgRTC cfg
+
         cBinOps = [ (Plus, "+"),  (Times, "*"), (Minus, "-")
                   , (Equal, "=="), (NotEqual, "!="), (LessThan, "<"), (GreaterThan, ">"), (LessEq, "<="), (GreaterEq, ">=")
                   , (And, "&"), (Or, "|"), (XOr, "^")
                   ]
+
+        -- see if we can find a constant shift; makes the output way more readable
+        getShiftAmnt def [_, sw] = case sw `lookup` consts of
+                                    Just (CW _  (CWInteger i)) -> integer i
+                                    _                          -> def
+        getShiftAmnt def _       = def
+
         p :: Op -> [Doc] -> Doc
         p (ArrRead _)       _  = tbd "User specified arrays (ArrRead)"
         p (ArrEq _ _)       _  = tbd "User specified arrays (ArrEq)"
         p (Label s)        [a] = a <+> text "/*" <+> text s <+> text "*/"
-        p (IEEEFP w)        as = handleIEEE w consts (zip opArgs as) var
+        p (IEEEFP w)         as = handleIEEE w  consts (zip opArgs as) var
+        p (PseudoBoolean pb) as = handlePB pb as
         p (KindCast _ to)   [a] = parens (text (show to)) <+> a
         p (Uninterpreted s) [] = text "/* Uninterpreted constant */" <+> text s
         p (Uninterpreted s) as = text "/* Uninterpreted function */" <+> text s <> parens (fsep (punctuate comma as))
@@ -640,8 +681,8 @@ ppExpr cfg consts (SBVApp op opArgs) lhs (typ, var)
         p Join [a, b]          = join (let (s1 : s2 : _) = opArgs in (s1, s2, a, b))
         p (Rol i) [a]          = rotate True  i a (head opArgs)
         p (Ror i) [a]          = rotate False i a (head opArgs)
-        p (Shl i) [a]          = shift  True  i a (head opArgs)
-        p (Shr i) [a]          = shift  False i a (head opArgs)
+        p Shl     [a, i]       = shift  True  (getShiftAmnt i opArgs) a -- The order of i/a being reversed here is
+        p Shr     [a, i]       = shift  False (getShiftAmnt i opArgs) a -- intentional and historical (from the days when Shl/Shr had a constant parameter.)
         p Not [a]              = case kindOf (head opArgs) of
                                    -- be careful about booleans, bitwise complement is not correct for them!
                                    KBool -> text "!" <> a
@@ -690,7 +731,22 @@ ppExpr cfg consts (SBVApp op opArgs) lhs (typ, var)
         p o [a, b]
           | Just co <- lookup o cBinOps
           = a <+> text co <+> b
+        p NotEqual xs = mkDistinct xs
         p o args = die $ "Received operator " ++ show o ++ " applied to " ++ show args
+
+        -- generate a pairwise inequality check
+        mkDistinct args = fsep $ andAll $ walk args
+          where walk []     = []
+                walk (e:es) = map (pair e) es ++ walk es
+
+                pair e1 e2  = parens (e1 <+> text "!=" <+> e2)
+
+                -- like punctuate, but more spacing
+                andAll []     = []
+                andAll (d:ds) = go d ds
+                     where go d' [] = [d']
+                           go d' (e:es) = (d' <+> text "&&") : go e es
+
         -- Div0 needs to protect, but only when the arguments are not float/double. (Div by 0 for those are well defined to be Inf/NaN etc.)
         protectDiv0 k divOp def a b = case k of
                                         KFloat  -> res
@@ -698,15 +754,11 @@ ppExpr cfg consts (SBVApp op opArgs) lhs (typ, var)
                                         _       -> wrap
            where res  = a <+> text divOp <+> b
                  wrap = parens (b <+> text "== 0") <+> text "?" <+> def <+> text ":" <+> parens res
-        shift toLeft i a s
-          | i < 0   = shift (not toLeft) (-i) a s
-          | i == 0  = a
-          | True    = case kindOf s of
-                        KBounded _ sz | i >= sz -> mkConst cfg $ mkConstCW (kindOf s) (0::Integer)
-                        KReal                   -> tbd $ "Shift for real quantity: " ++ show (toLeft, i, s)
-                        _                       -> a <+> text cop <+> int i
+
+        shift toLeft i a = a <+> text cop <+> i
           where cop | toLeft = "<<"
                     | True   = ">>"
+
         rotate toLeft i a s
           | i < 0   = rotate (not toLeft) (-i) a s
           | i == 0  = a
@@ -716,10 +768,11 @@ ppExpr cfg consts (SBVApp op opArgs) lhs (typ, var)
                         KBounded False sz           ->     parens (a <+> text cop  <+> int i)
                                                       <+> text "|"
                                                       <+> parens (a <+> text cop' <+> int (sz - i))
-                        KUnbounded                  -> shift toLeft i a s -- For SInteger, rotate is the same as shift in Haskell
+                        KUnbounded                  -> shift toLeft (int i) a -- For SInteger, rotate is the same as shift in Haskell
                         _                           -> tbd $ "Rotation for unbounded quantity: " ++ show (toLeft, i, s)
           where (cop, cop') | toLeft = ("<<", ">>")
                             | True   = (">>", "<<")
+
         -- TBD: below we only support the values for extract that are "easy" to implement. These should cover
         -- almost all instances actually generated by SBV, however.
         extract hi lo i a  -- Isolate the bit-extraction case
@@ -741,6 +794,7 @@ ppExpr cfg consts (SBVApp op opArgs) lhs (typ, var)
                               ( 7,  0, KBounded False  8) -> text "(SInt8)"   <+> a
                               ( 7,  0, KBounded True   8) -> text "(SWord8)"  <+> a
                               ( _,  _, k                ) -> tbd $ "extract with " ++ show (hi, lo, k, i)
+
         -- TBD: ditto here for join, just like extract above
         join (i, j, a, b) = case (kindOf i, kindOf j) of
                               (KBounded False  8, KBounded False  8) -> parens (parens (text "(SWord16)" <+> a) <+> text "<< 8")  <+> text "|" <+> parens (text "(SWord16)" <+> b)
