@@ -12,11 +12,37 @@
 {-# LANGUAGE FlexibleInstances          #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
 
-module Data.SBV.Compilers.CodeGen where
+module Data.SBV.Compilers.CodeGen (
+        -- * The codegen monad
+          SBVCodeGen(..)
+
+        -- * Specifying inputs, SBV variants
+        , cgInput,  cgInputArr
+        , cgOutput, cgOutputArr
+        , cgReturn, cgReturnArr
+
+        -- * Specifying inputs, SVal variants
+        , svCgInput,  svCgInputArr
+        , svCgOutput, svCgOutputArr
+        , svCgReturn, svCgReturnArr
+
+        -- * Settings
+        , cgPerformRTCs, cgSetDriverValues
+        , cgAddPrototype, cgAddDecl, cgAddLDFlags, cgIgnoreSAssert
+        , cgIntegerSize, cgSRealType, CgSRealType(..)
+
+        -- * Infrastructure
+        , CgTarget(..), CgConfig(..), CgState(..), CgPgmBundle(..), CgPgmKind(..), CgVal(..)
+        , defaultCgConfig, initCgState, isCgDriver, isCgMakefile
+
+        -- * Generating collateral
+        , cgGenerateDriver, cgGenerateMakefile, codeGen, renderCgPgmBundle
+        ) where
 
 import Control.Monad             (filterM, replicateM, unless)
-import Control.Monad.Trans
-import Control.Monad.State.Lazy  (MonadState, StateT(..), modify)
+import Control.Monad.Reader      (ask)
+import Control.Monad.Trans       (MonadIO, lift, liftIO)
+import Control.Monad.State.Lazy  (MonadState, StateT(..), modify')
 import Data.Char                 (toLower, isSpace)
 import Data.List                 (nub, isPrefixOf, intercalate, (\\))
 import System.Directory          (createDirectoryIfMissing, doesDirectoryExist, doesFileExist)
@@ -26,11 +52,8 @@ import System.IO                 (hFlush, stdout)
 import           Text.PrettyPrint.HughesPJ      (Doc, vcat)
 import qualified Text.PrettyPrint.HughesPJ as P (render)
 
-import Data.SBV.BitVectors.Data
-import Data.SBV.BitVectors.Symbolic (svToSymSW, svMkSymVar, outputSVal)
-
-import Prelude ()
-import Prelude.Compat
+import Data.SBV.Core.Data
+import Data.SBV.Core.Symbolic (svToSymSW, svMkSymVar, outputSVal)
 
 -- | Abstract over code generation for different languages
 class CgTarget a where
@@ -102,7 +125,7 @@ cgSBVToSW = liftSymbolic . sbvToSymSW
 
 -- | Sets RTC (run-time-checks) for index-out-of-bounds, shift-with-large value etc. on/off. Default: 'False'.
 cgPerformRTCs :: Bool -> SBVCodeGen ()
-cgPerformRTCs b = modify (\s -> s { cgFinalConfig = (cgFinalConfig s) { cgRTC = b } })
+cgPerformRTCs b = modify' (\s -> s { cgFinalConfig = (cgFinalConfig s) { cgRTC = b } })
 
 -- | Sets number of bits to be used for representing the 'SInteger' type in the generated C code.
 -- The argument must be one of @8@, @16@, @32@, or @64@. Note that this is essentially unsafe as
@@ -113,7 +136,7 @@ cgIntegerSize i
   | i `notElem` [8, 16, 32, 64]
   = error $ "SBV.cgIntegerSize: Argument must be one of 8, 16, 32, or 64. Received: " ++ show i
   | True
-  = modify (\s -> s { cgFinalConfig = (cgFinalConfig s) { cgInteger = Just i }})
+  = modify' (\s -> s { cgFinalConfig = (cgFinalConfig s) { cgInteger = Just i }})
 
 -- | Possible mappings for the 'SReal' type when translated to C. Used in conjunction
 -- with the function 'cgSRealType'. Note that the particular characteristics of the
@@ -124,7 +147,7 @@ data CgSRealType = CgFloat      -- ^ @float@
                  | CgLongDouble -- ^ @long double@
                  deriving Eq
 
--- | 'Show' instance for 'cgSRealType' displays values as they would be used in a C program
+-- 'Show' instance for 'cgSRealType' displays values as they would be used in a C program
 instance Show CgSRealType where
   show CgFloat      = "float"
   show CgDouble     = "double"
@@ -136,60 +159,60 @@ instance Show CgSRealType where
 -- infinite precision SReal values becomes reduced to the corresponding floating point type in
 -- C, and hence it is subject to rounding errors.
 cgSRealType :: CgSRealType -> SBVCodeGen ()
-cgSRealType rt = modify (\s -> s {cgFinalConfig = (cgFinalConfig s) { cgReal = Just rt }})
+cgSRealType rt = modify' (\s -> s {cgFinalConfig = (cgFinalConfig s) { cgReal = Just rt }})
 
 -- | Should we generate a driver program? Default: 'True'. When a library is generated, it will have
 -- a driver if any of the contituent functions has a driver. (See 'compileToCLib'.)
 cgGenerateDriver :: Bool -> SBVCodeGen ()
-cgGenerateDriver b = modify (\s -> s { cgFinalConfig = (cgFinalConfig s) { cgGenDriver = b } })
+cgGenerateDriver b = modify' (\s -> s { cgFinalConfig = (cgFinalConfig s) { cgGenDriver = b } })
 
 -- | Should we generate a Makefile? Default: 'True'.
 cgGenerateMakefile :: Bool -> SBVCodeGen ()
-cgGenerateMakefile b = modify (\s -> s { cgFinalConfig = (cgFinalConfig s) { cgGenMakefile = b } })
+cgGenerateMakefile b = modify' (\s -> s { cgFinalConfig = (cgFinalConfig s) { cgGenMakefile = b } })
 
 -- | Sets driver program run time values, useful for generating programs with fixed drivers for testing. Default: None, i.e., use random values.
 cgSetDriverValues :: [Integer] -> SBVCodeGen ()
-cgSetDriverValues vs = modify (\s -> s { cgFinalConfig = (cgFinalConfig s) { cgDriverVals = vs } })
+cgSetDriverValues vs = modify' (\s -> s { cgFinalConfig = (cgFinalConfig s) { cgDriverVals = vs } })
 
 -- | Ignore assertions (those generated by 'sAssert' calls) in the generated C code
 cgIgnoreSAssert :: Bool -> SBVCodeGen ()
-cgIgnoreSAssert b = modify (\s -> s { cgFinalConfig = (cgFinalConfig s) { cgIgnoreAsserts = b } })
+cgIgnoreSAssert b = modify' (\s -> s { cgFinalConfig = (cgFinalConfig s) { cgIgnoreAsserts = b } })
 
 -- | Adds the given lines to the header file generated, useful for generating programs with uninterpreted functions.
 cgAddPrototype :: [String] -> SBVCodeGen ()
-cgAddPrototype ss = modify (\s -> let old = cgPrototypes s
-                                      new = if null old then ss else old ++ [""] ++ ss
-                                  in s { cgPrototypes = new })
+cgAddPrototype ss = modify' (\s -> let old = cgPrototypes s
+                                       new = if null old then ss else old ++ [""] ++ ss
+                                   in s { cgPrototypes = new })
 
 -- | Adds the given lines to the program file generated, useful for generating programs with uninterpreted functions.
 cgAddDecl :: [String] -> SBVCodeGen ()
-cgAddDecl ss = modify (\s -> s { cgDecls = cgDecls s ++ ss })
+cgAddDecl ss = modify' (\s -> s { cgDecls = cgDecls s ++ ss })
 
 -- | Adds the given words to the compiler options in the generated Makefile, useful for linking extra stuff in.
 cgAddLDFlags :: [String] -> SBVCodeGen ()
-cgAddLDFlags ss = modify (\s -> s { cgLDFlags = cgLDFlags s ++ ss })
+cgAddLDFlags ss = modify' (\s -> s { cgLDFlags = cgLDFlags s ++ ss })
 
 -- | Creates an atomic input in the generated code.
 svCgInput :: Kind -> String -> SBVCodeGen SVal
-svCgInput k nm = do r <- liftSymbolic (svMkSymVar (Just ALL) k Nothing)
+svCgInput k nm = do r  <- liftSymbolic (ask >>= liftIO . svMkSymVar (Just ALL) k Nothing)
                     sw <- liftSymbolic (svToSymSW r)
-                    modify (\s -> s { cgInputs = (nm, CgAtomic sw) : cgInputs s })
+                    modify' (\s -> s { cgInputs = (nm, CgAtomic sw) : cgInputs s })
                     return r
 
 -- | Creates an array input in the generated code.
 svCgInputArr :: Kind -> Int -> String -> SBVCodeGen [SVal]
 svCgInputArr k sz nm
   | sz < 1 = error $ "SBV.cgInputArr: Array inputs must have at least one element, given " ++ show sz ++ " for " ++ show nm
-  | True   = do rs <- liftSymbolic $ replicateM sz (svMkSymVar (Just ALL) k Nothing)
+  | True   = do rs  <- liftSymbolic $ ask >>= liftIO . replicateM sz . svMkSymVar (Just ALL) k Nothing
                 sws <- liftSymbolic $ mapM svToSymSW rs
-                modify (\s -> s { cgInputs = (nm, CgArray sws) : cgInputs s })
+                modify' (\s -> s { cgInputs = (nm, CgArray sws) : cgInputs s })
                 return rs
 
 -- | Creates an atomic output in the generated code.
 svCgOutput :: String -> SVal -> SBVCodeGen ()
 svCgOutput nm v = do _ <- liftSymbolic (outputSVal v)
                      sw <- liftSymbolic (svToSymSW v)
-                     modify (\s -> s { cgOutputs = (nm, CgAtomic sw) : cgOutputs s })
+                     modify' (\s -> s { cgOutputs = (nm, CgAtomic sw) : cgOutputs s })
 
 -- | Creates an array output in the generated code.
 svCgOutputArr :: String -> [SVal] -> SBVCodeGen ()
@@ -197,14 +220,14 @@ svCgOutputArr nm vs
   | sz < 1 = error $ "SBV.cgOutputArr: Array outputs must have at least one element, received " ++ show sz ++ " for " ++ show nm
   | True   = do _ <- liftSymbolic (mapM outputSVal vs)
                 sws <- liftSymbolic (mapM svToSymSW vs)
-                modify (\s -> s { cgOutputs = (nm, CgArray sws) : cgOutputs s })
+                modify' (\s -> s { cgOutputs = (nm, CgArray sws) : cgOutputs s })
   where sz = length vs
 
 -- | Creates a returned (unnamed) value in the generated code.
 svCgReturn :: SVal -> SBVCodeGen ()
 svCgReturn v = do _ <- liftSymbolic (outputSVal v)
                   sw <- liftSymbolic (svToSymSW v)
-                  modify (\s -> s { cgReturns = CgAtomic sw : cgReturns s })
+                  modify' (\s -> s { cgReturns = CgAtomic sw : cgReturns s })
 
 -- | Creates a returned (unnamed) array value in the generated code.
 svCgReturnArr :: [SVal] -> SBVCodeGen ()
@@ -212,14 +235,14 @@ svCgReturnArr vs
   | sz < 1 = error $ "SBV.cgReturnArr: Array returns must have at least one element, received " ++ show sz
   | True   = do _ <- liftSymbolic (mapM outputSVal vs)
                 sws <- liftSymbolic (mapM svToSymSW vs)
-                modify (\s -> s { cgReturns = CgArray sws : cgReturns s })
+                modify' (\s -> s { cgReturns = CgArray sws : cgReturns s })
   where sz = length vs
 
 -- | Creates an atomic input in the generated code.
 cgInput :: SymWord a => String -> SBVCodeGen (SBV a)
 cgInput nm = do r <- liftSymbolic forall_
                 sw <- cgSBVToSW r
-                modify (\s -> s { cgInputs = (nm, CgAtomic sw) : cgInputs s })
+                modify' (\s -> s { cgInputs = (nm, CgAtomic sw) : cgInputs s })
                 return r
 
 -- | Creates an array input in the generated code.
@@ -228,14 +251,14 @@ cgInputArr sz nm
   | sz < 1 = error $ "SBV.cgInputArr: Array inputs must have at least one element, given " ++ show sz ++ " for " ++ show nm
   | True   = do rs <- liftSymbolic $ mapM (const forall_) [1..sz]
                 sws <- mapM cgSBVToSW rs
-                modify (\s -> s { cgInputs = (nm, CgArray sws) : cgInputs s })
+                modify' (\s -> s { cgInputs = (nm, CgArray sws) : cgInputs s })
                 return rs
 
 -- | Creates an atomic output in the generated code.
 cgOutput :: String -> SBV a -> SBVCodeGen ()
 cgOutput nm v = do _ <- liftSymbolic (output v)
                    sw <- cgSBVToSW v
-                   modify (\s -> s { cgOutputs = (nm, CgAtomic sw) : cgOutputs s })
+                   modify' (\s -> s { cgOutputs = (nm, CgAtomic sw) : cgOutputs s })
 
 -- | Creates an array output in the generated code.
 cgOutputArr :: SymWord a => String -> [SBV a] -> SBVCodeGen ()
@@ -243,14 +266,14 @@ cgOutputArr nm vs
   | sz < 1 = error $ "SBV.cgOutputArr: Array outputs must have at least one element, received " ++ show sz ++ " for " ++ show nm
   | True   = do _ <- liftSymbolic (mapM output vs)
                 sws <- mapM cgSBVToSW vs
-                modify (\s -> s { cgOutputs = (nm, CgArray sws) : cgOutputs s })
+                modify' (\s -> s { cgOutputs = (nm, CgArray sws) : cgOutputs s })
   where sz = length vs
 
 -- | Creates a returned (unnamed) value in the generated code.
 cgReturn :: SBV a -> SBVCodeGen ()
 cgReturn v = do _ <- liftSymbolic (output v)
                 sw <- cgSBVToSW v
-                modify (\s -> s { cgReturns = CgAtomic sw : cgReturns s })
+                modify' (\s -> s { cgReturns = CgAtomic sw : cgReturns s })
 
 -- | Creates a returned (unnamed) array value in the generated code.
 cgReturnArr :: SymWord a => [SBV a] -> SBVCodeGen ()
@@ -258,7 +281,7 @@ cgReturnArr vs
   | sz < 1 = error $ "SBV.cgReturnArr: Array returns must have at least one element, received " ++ show sz
   | True   = do _ <- liftSymbolic (mapM output vs)
                 sws <- mapM cgSBVToSW vs
-                modify (\s -> s { cgReturns = CgArray sws : cgReturns s })
+                modify' (\s -> s { cgReturns = CgArray sws : cgReturns s })
   where sz = length vs
 
 -- | Representation of a collection of generated programs.
@@ -280,7 +303,7 @@ isCgMakefile :: CgPgmKind -> Bool
 isCgMakefile CgMakefile{} = True
 isCgMakefile _            = False
 
--- | A simple way to print bundles, mostly for debugging purposes.
+-- A simple way to print bundles, mostly for debugging purposes.
 instance Show CgPgmBundle where
    show (CgPgmBundle _ fs) = intercalate "\n" $ map showFile fs
     where showFile :: (FilePath, (CgPgmKind, [Doc])) -> String
@@ -292,7 +315,7 @@ instance Show CgPgmBundle where
 -- of makefiles, source code, headers, etc.
 codeGen :: CgTarget l => l -> CgConfig -> String -> SBVCodeGen () -> IO CgPgmBundle
 codeGen l cgConfig nm (SBVCodeGen comp) = do
-   (((), st'), res) <- runSymbolic' CodeGen $ runStateT comp initCgState { cgFinalConfig = cgConfig }
+   (((), st'), res) <- runSymbolic CodeGen $ runStateT comp initCgState { cgFinalConfig = cgConfig }
    let st = st' { cgInputs       = reverse (cgInputs st')
                 , cgOutputs      = reverse (cgOutputs st')
                 }
