@@ -6,13 +6,14 @@
 -- Maintainer  :  erkokl@gmail.com
 -- Stability   :  experimental
 --
--- Computing population-counts (number of set bits) and autimatically
+-- Computing population-counts (number of set bits) and automatically
 -- generating C code.
 -----------------------------------------------------------------------------
 
 module Data.SBV.Examples.CodeGeneration.PopulationCount where
 
 import Data.SBV
+import Data.SBV.Tools.CodeGen
 
 -----------------------------------------------------------------------------
 -- * Reference: Slow but /obviously/ correct
@@ -63,10 +64,12 @@ respect to the slower reference version.
 -}
 
 -- | States the correctness of faster population-count algorithm, with respect
--- to the reference slow version. (We use yices here as it's quite fast for
--- this problem. Z3 seems to take much longer.) We have:
+-- to the reference slow version. Turns out Z3's default solver is rather slow
+-- for this one, but there's a magic incantation to make it go fast.
+-- See <https://github.com/Z3Prover/z3/issues/1150> for details.
 --
--- >>> proveWith yices fastPopCountIsCorrect
+-- >>> let cmd = "(check-sat-using (then (using-params ackermannize_bv :div0_ackermann_limit 1000000) simplify bit-blast sat))"
+-- >>> proveWith z3{satCmd = cmd} fastPopCountIsCorrect
 -- Q.E.D.
 fastPopCountIsCorrect :: SWord64 -> SBool
 fastPopCountIsCorrect x = popCountFast x .== popCountSlow x
@@ -188,36 +191,36 @@ fastPopCountIsCorrect x = popCountFast x .== popCountSlow x
 --   };
 --   const SWord64 s11 = s0 & 0x00000000000000ffULL;
 --   const SWord8  s12 = table0[s11];
---   const SWord64 s13 = s0 >> 8;
---   const SWord64 s14 = 0x00000000000000ffULL & s13;
---   const SWord8  s15 = table0[s14];
---   const SWord8  s16 = s12 + s15;
---   const SWord64 s17 = s13 >> 8;
---   const SWord64 s18 = 0x00000000000000ffULL & s17;
---   const SWord8  s19 = table0[s18];
---   const SWord8  s20 = s16 + s19;
---   const SWord64 s21 = s17 >> 8;
---   const SWord64 s22 = 0x00000000000000ffULL & s21;
---   const SWord8  s23 = table0[s22];
---   const SWord8  s24 = s20 + s23;
---   const SWord64 s25 = s21 >> 8;
---   const SWord64 s26 = 0x00000000000000ffULL & s25;
---   const SWord8  s27 = table0[s26];
---   const SWord8  s28 = s24 + s27;
---   const SWord64 s29 = s25 >> 8;
---   const SWord64 s30 = 0x00000000000000ffULL & s29;
---   const SWord8  s31 = table0[s30];
---   const SWord8  s32 = s28 + s31;
---   const SWord64 s33 = s29 >> 8;
---   const SWord64 s34 = 0x00000000000000ffULL & s33;
---   const SWord8  s35 = table0[s34];
---   const SWord8  s36 = s32 + s35;
---   const SWord64 s37 = s33 >> 8;
---   const SWord64 s38 = 0x00000000000000ffULL & s37;
---   const SWord8  s39 = table0[s38];
---   const SWord8  s40 = s36 + s39;
+--   const SWord64 s14 = s0 >> 8;
+--   const SWord64 s15 = 0x00000000000000ffULL & s14;
+--   const SWord8  s16 = table0[s15];
+--   const SWord8  s17 = s12 + s16;
+--   const SWord64 s18 = s14 >> 8;
+--   const SWord64 s19 = 0x00000000000000ffULL & s18;
+--   const SWord8  s20 = table0[s19];
+--   const SWord8  s21 = s17 + s20;
+--   const SWord64 s22 = s18 >> 8;
+--   const SWord64 s23 = 0x00000000000000ffULL & s22;
+--   const SWord8  s24 = table0[s23];
+--   const SWord8  s25 = s21 + s24;
+--   const SWord64 s26 = s22 >> 8;
+--   const SWord64 s27 = 0x00000000000000ffULL & s26;
+--   const SWord8  s28 = table0[s27];
+--   const SWord8  s29 = s25 + s28;
+--   const SWord64 s30 = s26 >> 8;
+--   const SWord64 s31 = 0x00000000000000ffULL & s30;
+--   const SWord8  s32 = table0[s31];
+--   const SWord8  s33 = s29 + s32;
+--   const SWord64 s34 = s30 >> 8;
+--   const SWord64 s35 = 0x00000000000000ffULL & s34;
+--   const SWord8  s36 = table0[s35];
+--   const SWord8  s37 = s33 + s36;
+--   const SWord64 s38 = s34 >> 8;
+--   const SWord64 s39 = 0x00000000000000ffULL & s38;
+--   const SWord8  s40 = table0[s39];
+--   const SWord8  s41 = s37 + s40;
 -- <BLANKLINE>
---   return s40;
+--   return s41;
 -- }
 -- == END: "popCount.c" ==================
 genPopCountInC :: IO ()
