@@ -1,4 +1,4 @@
-----------------------------------------------------------------------------
+-----------------------------------------------------------------------------
 -- |
 -- Module      :  Data.SBV.SMT.SMTLib2
 -- Copyright   :  (c) Levent Erkok
@@ -10,116 +10,67 @@
 -----------------------------------------------------------------------------
 {-# LANGUAGE PatternGuards #-}
 
-module Data.SBV.SMT.SMTLib2(cvt, addNonEqConstraints) where
+module Data.SBV.SMT.SMTLib2(cvt, cvtInc) where
 
-import Data.Bits     (bit)
-import Data.Function (on)
-import Data.Ord      (comparing)
-import Data.List     (intercalate, partition, groupBy, sortBy)
+import Data.Bits  (bit)
+import Data.List  (intercalate, partition)
 
 import qualified Data.Foldable as F (toList)
 import qualified Data.Map      as M
 import qualified Data.IntMap   as IM
 import qualified Data.Set      as Set
 
-import Data.SBV.BitVectors.Data
-import Data.SBV.BitVectors.PrettyNum (smtRoundingMode, cwToSMTLib)
+import Data.SBV.Core.Data
+import Data.SBV.SMT.Utils
+import Data.SBV.Control.Types
 
--- | Add constraints to generate /new/ models. This function is used to query the SMT-solver, while
--- disallowing a previous model.
-addNonEqConstraints :: RoundingMode -> [(Quantifier, NamedSymVar)] -> [[(String, CW)]] -> SMTLibPgm -> Maybe String
-addNonEqConstraints rm qinps allNonEqConstraints (SMTLibPgm _ (aliasTable, pre, post))
-  | null allNonEqConstraints
-  = Just $ intercalate "\n" $ pre ++ post
-  | null refutedModel
-  = Nothing
-  | True
-  = Just $ intercalate "\n" $ pre
-    ++ [ "; --- refuted-models ---" ]
-    ++ refutedModel
-    ++ post
- where refutedModel = concatMap (nonEqs rm . map intName) nonEqConstraints
-       intName (s, c)
-          | Just sw <- s `lookup` aliasTable = (show sw, c)
-          | True                             = (s, c)
-       -- with existentials, we only add top-level existentials to the refuted-models list
-       nonEqConstraints = filter (not . null) $ map (filter (\(s, _) -> s `elem` topUnivs)) allNonEqConstraints
-       topUnivs = [s | (_, (_, s)) <- takeWhile (\p -> fst p == EX) qinps]
-
-nonEqs :: RoundingMode -> [(String, CW)] -> [String]
-nonEqs rm scs = format $ interp ps ++ disallow (map eqClass uninterpClasses)
-  where isFree (KUserSort _ (Left _)) = True
-        isFree _                      = False
-        (ups, ps) = partition (isFree . kindOf . snd) scs
-        format []     =  []
-        format [m]    =  ["(assert " ++ m ++ ")"]
-        format (m:ms) =  ["(assert (or " ++ m]
-                      ++ map ("            " ++) ms
-                      ++ ["        ))"]
-        -- Regular (or interpreted) sorts simply get a constraint that we disallow the current assignment
-        interp = map $ nonEq rm
-        -- Determine the equivalence classes of uninterpreted sorts:
-        uninterpClasses = filter (\l -> length l > 1) -- Only need this class if it has at least two members
-                        . map (map fst)               -- throw away sorts, we only need the names
-                        . groupBy ((==) `on` snd)     -- make sure they belong to the same sort and have the same value
-                        . sortBy (comparing snd)      -- sort them according to their sorts first
-                        $ ups                         -- take the uninterpreted sorts
-        -- Uninterpreted sorts get a constraint that says the equivalence classes as determined by the solver are disallowed:
-        eqClass :: [String] -> String
-        eqClass [] = error "SBV.allSat.nonEqs: Impossible happened, disallow received an empty list"
-        eqClass cs = "(= " ++ unwords cs ++ ")"
-        -- Now, take the conjunction of equivalence classes and assert it's negation:
-        disallow = map $ \ec -> "(not " ++ ec ++ ")"
-
-nonEq :: RoundingMode -> (String, CW) -> String
-nonEq rm (s, c) = "(not (= " ++ s ++ " " ++ cvtCW rm c ++ "))"
+import Data.SBV.Utils.PrettyNum (smtRoundingMode, cwToSMTLib)
 
 tbd :: String -> a
 tbd e = error $ "SBV.SMTLib2: Not-yet-supported: " ++ e
 
 -- | Translate a problem into an SMTLib2 script
-cvt :: RoundingMode                 -- ^ User selected rounding mode to be used for floating point arithmetic
-    -> Maybe Logic                  -- ^ SMT-Lib logic, if requested by the user
-    -> SolverCapabilities           -- ^ capabilities of the current solver
-    -> Set.Set Kind                 -- ^ kinds used
-    -> Bool                         -- ^ is this a sat problem?
-    -> [String]                     -- ^ extra comments to place on top
-    -> [(Quantifier, NamedSymVar)]  -- ^ inputs
-    -> [Either SW (SW, [SW])]       -- ^ skolemized version inputs
-    -> [(SW, CW)]                   -- ^ constants
-    -> [((Int, Kind, Kind), [SW])]  -- ^ auto-generated tables
-    -> [(Int, ArrayInfo)]           -- ^ user specified arrays
-    -> [(String, SBVType)]          -- ^ uninterpreted functions/constants
-    -> [(String, [String])]         -- ^ user given axioms
-    -> SBVPgm                       -- ^ assignments
-    -> [SW]                         -- ^ extra constraints
-    -> SW                           -- ^ output variable
-    -> ([String], [String])
-cvt rm smtLogic solverCaps kindInfo isSat comments inputs skolemInps consts tbls arrs uis axs (SBVPgm asgnsSeq) cstrs out = (pre, [])
-  where -- the logic is an over-approaximation
-        hasInteger     = KUnbounded `Set.member` kindInfo
+cvt :: SMTLibConverter [String]
+cvt kindInfo isSat comments (inputs, trackerVars) skolemInps consts tbls arrs uis axs (SBVPgm asgnsSeq) cstrs out cfg = pgm
+  where hasInteger     = KUnbounded `Set.member` kindInfo
         hasReal        = KReal      `Set.member` kindInfo
         hasFloat       = KFloat     `Set.member` kindInfo
         hasDouble      = KDouble    `Set.member` kindInfo
         hasBVs         = not $ null [() | KBounded{} <- Set.toList kindInfo]
         usorts         = [(s, dt) | KUserSort s dt <- Set.toList kindInfo]
         hasNonBVArrays = (not . null) [() | (_, (_, (k1, k2), _)) <- arrs, not (isBounded k1 && isBounded k2)]
+        rm             = roundingMode cfg
+        solverCaps     = capabilities (solver cfg)
+
+        -- Determining the logic is surprisingly tricky!
         logic
-           | Just l <- smtLogic
-           = ["(set-logic " ++ show l ++ ") ; NB. User specified."]
-           | hasDouble || hasFloat    -- NB. We don't check for quantifiers here, we probably should..
-           = if hasBVs
-             then ["(set-logic QF_FPBV)"]
-             else ["(set-logic QF_FP)"]
+           -- user told us what to do: so just take it:
+           | Just l <- case [l | SetLogic l <- solverSetOptions cfg] of
+                         []  -> Nothing
+                         [l] -> Just l
+                         ls  -> error $ unlines [ ""
+                                                , "*** Only one setOption call to 'SetLogic' is allowed, found: " ++ show (length ls)
+                                                , "***  " ++ unwords (map show ls)
+                                                ]
+           = case l of
+               Logic_NONE -> ["; NB. Not setting the logic per user request of Logic_NONE"]
+               _          -> ["(set-logic " ++ show l ++ ") ; NB. User specified."]
+
+           -- Otherwise, we try to determine the most suitable logic.
+           -- NB. This isn't really fool proof!
+           | hasDouble || hasFloat
+           = if hasInteger || not (null foralls)
+             then ["(set-logic ALL)"]
+             else if hasBVs
+                  then ["(set-logic QF_FPBV)"]
+                  else ["(set-logic QF_FP)"]
            | hasInteger || hasReal || not (null usorts) || hasNonBVArrays
            = let why | hasInteger        = "has unbounded values"
                      | hasReal           = "has algebraic reals"
                      | not (null usorts) = "has user-defined sorts"
                      | hasNonBVArrays    = "has non-bitvector arrays"
                      | True              = "cannot determine the SMTLib-logic to use"
-             in case mbDefaultLogic solverCaps hasReal of
-                  Nothing -> ["; " ++ why ++ ", no logic specified."]
-                  Just l  -> ["(set-logic " ++ l ++ "); " ++ why ++ ", using solver-default logic."]
+             in ["(set-logic ALL) ; "  ++ why ++ ", using catch-all."]
            | True
            = ["(set-logic " ++ qs ++ as ++ ufs ++ "BV)"]
           where qs  | null foralls && null axs = "QF_"  -- axioms are likely to contain quantifiers
@@ -128,19 +79,29 @@ cvt rm smtLogic solverCaps kindInfo isSat comments inputs skolemInps consts tbls
                     | True                     = "A"
                 ufs | null uis && null tbls    = ""     -- we represent tables as UFs
                     | True                     = "UF"
-        getModels
-          | supportsProduceModels solverCaps = ["(set-option :produce-models true)"]
-          | True                             = []
-        pre  =  ["; Automatically generated by SBV. Do not edit."]
-             ++ map ("; " ++) comments
-             ++ getModels
-             ++ logic
+
+        -- SBV always requires the production of models!
+        getModels   = ["(set-option :produce-models true)"]
+
+        -- process all other settings we're given
+        userSettings = concatMap opts $ solverSetOptions cfg
+           where opts SetLogic{} = []     -- processed already
+                 opts o          = [setSMTOption o]
+
+        settings =  userSettings        -- NB. Make sure this comes first!
+                 ++ getModels
+                 ++ logic
+
+        pgm  =  map ("; " ++) comments
+             ++ settings
              ++ [ "; --- uninterpreted sorts ---" ]
              ++ concatMap declSort usorts
              ++ [ "; --- literal constants ---" ]
-             ++ concatMap (declConst (supportsMacros solverCaps)) consts
+             ++ map (declConst cfg) consts
              ++ [ "; --- skolem constants ---" ]
              ++ [ "(declare-fun " ++ show s ++ " " ++ swFunType ss s ++ ")" ++ userName s | Right (s, ss) <- skolemInps]
+             ++ [ "; --- optimization tracker variables ---" | not (null trackerVars) ]
+             ++ [ "(declare-fun " ++ show s ++ " " ++ swFunType [] s ++ ") ; tracks " ++ nm | (s, nm) <- trackerVars]
              ++ [ "; --- constant tables ---" ]
              ++ concatMap constTable constTables
              ++ [ "; --- skolemized tables ---" ]
@@ -152,62 +113,182 @@ cvt rm smtLogic solverCaps kindInfo isSat comments inputs skolemInps consts tbls
              ++ [ "; --- user given axioms ---" ]
              ++ map declAx axs
              ++ [ "; --- formula ---" ]
-             ++ [if null foralls
-                 then "(assert ; no quantifiers"
-                 else "(assert (forall (" ++ intercalate "\n                 "
-                                             ["(" ++ show s ++ " " ++ swType s ++ ")" | s <- foralls] ++ ")"]
-             ++ map (letAlign . mkLet) asgns
-             ++ map letAlign (if null delayedEqualities then [] else ("(and " ++ deH) : map (align 5) deTs)
-             ++ [ impAlign (letAlign assertOut) ++ replicate noOfCloseParens ')' ]
-        noOfCloseParens = length asgns + (if null foralls then 1 else 2) + (if null delayedEqualities then 0 else 1)
+
+             ++ map (declDef cfg skolemMap tableMap) preQuantifierAssigns
+             ++ ["(assert (forall (" ++ intercalate "\n                 "
+                                        ["(" ++ show s ++ " " ++ swType s ++ ")" | s <- foralls] ++ ")"
+                | not (null foralls)
+                ]
+             ++ map mkAssign postQuantifierAssigns
+
+             ++ delayedAsserts delayedEqualities
+
+             ++ finalAssert
+
+        -- identify the assignments that can come before the first quantifier
+        (preQuantifierAssigns, postQuantifierAssigns)
+           | null foralls
+           = ([], asgns)  -- the apparent "switch" here is OK; rest of the code works correctly if there are no foralls.
+           | True
+           = span pre asgns
+           where first      = nodeId (minimum foralls)
+                 pre (s, _) = nodeId s < first
+
+                 nodeId (SW _ n) = n
+
+        noOfCloseParens
+          | null foralls = 0
+          | True         = length postQuantifierAssigns + 2 + (if null delayedEqualities then 0 else 1)
+
+        foralls    = [s | Left s <- skolemInps]
+        forallArgs = concatMap ((" " ++) . show) foralls
+
         (constTables, skolemTables) = ([(t, d) | (t, Left d) <- allTables], [(t, d) | (t, Right d) <- allTables])
         allTables = [(t, genTableData rm skolemMap (not (null foralls), forallArgs) (map fst consts) t) | t <- tbls]
         (arrayConstants, allArrayDelayeds) = unzip $ map (declArray (not (null foralls)) (map fst consts) skolemMap) arrs
-        delayedEqualities@(~(deH:deTs)) = concatMap snd skolemTables ++ concat allArrayDelayeds
-        foralls = [s | Left s <- skolemInps]
-        forallArgs = concatMap ((" " ++) . show) foralls
-        letAlign s
-          | null foralls = "   " ++ s
-          | True         = "            " ++ s
+        delayedEqualities = concatMap snd skolemTables ++ concat allArrayDelayeds
+
+        delayedAsserts []              = []
+        delayedAsserts ds@(deH : deTs)
+          | null foralls = map (\s -> "(assert " ++ s ++ ")") ds
+          | True         = map letShift (("(and " ++ deH) : map (align 5) deTs)
+
+        letShift = align 12
+
+        finalAssert
+          | null foralls
+          = map (\a -> "(assert " ++ named a ++ ")") assertions
+          | not (null namedAsserts)
+          = error $ intercalate "\n" [ "SBV: Named constraints and quantifiers cannot be mixed!"
+                                     , "   Quantified variables: " ++ unwords (map show foralls)
+                                     , "   Named constraints   : " ++ intercalate ", " (map show namedAsserts)
+                                     ]
+          | True
+          = [impAlign (letShift combined) ++ replicate noOfCloseParens ')']
+          where namedAsserts = [n | (Just n, _) <- assertions]
+
+                combined = case map snd assertions of
+                             [x] -> x
+                             xs  -> "(and " ++ unwords xs ++ ")"
+
         impAlign s
           | null delayedEqualities = s
           | True                   = "     " ++ s
+
         align n s = replicate n ' ' ++ s
-        -- if sat,   we assert cstrs /\ out
-        -- if prove, we assert ~(cstrs => out) = cstrs /\ not out
-        assertOut
-           | null cstrs = o
-           | True       = "(and " ++ unwords (map mkConj cstrs ++ [o]) ++ ")"
-           where mkConj = cvtSW skolemMap
-                 o | isSat =            mkConj out
-                   | True  = "(not " ++ mkConj out ++ ")"
+
+        -- We have:
+        --    - cstrs   : Explicitly given constraints (via calls to constrain)
+        --    - p1..pn  : The path conditions in a case-split that led us here. This is given in a case-split.
+        --    - c1..cm  : All the other case-split constraints for the coverage case. This is in a case-coverage.
+        -- if sat:
+        --     -- we assert (cstrs /\ (p1 /\ p2 /\ ... /\ pn) /\ ~(c1 \/ c2 \/ .. \/ cm) /\ out)
+        --            i.e., cstrs /\ p1 /\ p2 /\ ... /\ pn /\ ~c1 /\ ~c2 /\ ~c3 .. /\ ~cm /\ out
+        -- if prove:
+        --     -- we assert ~((cstrs /\  (p1 /\ p2 /\ .. /\ pn) /\ ~(c1 \/ c2 \/ .. \/ cm)) => out)
+        --            i.e., cstrs /\ p1 /\ p2 /\ .. /\ pn /\ ~c1 /\ ~c2 /\ ~c3 .. /\ ~cm /\ ~out
+        -- That is, we always assert all path constraints and path conditions AND
+        --     -- negation of the output in a prove
+        --     -- output itself in a sat
+        assertions
+           | null finals = [(Nothing, cvtSW skolemMap trueSW)]
+           | True        = finals
+
+           where finals  = cstrs' ++ maybe [] (\r -> [(Nothing, r)]) mbO
+
+                 cstrs' =  [(mbNm, c') | (mbNm, c) <- cstrs, Just c' <- [pos c]]
+
+                 mbO | isSat = pos out
+                     | True  = neg out
+
+                 neg s
+                  | s == trueSW  = Just $ cvtSW skolemMap falseSW
+                  | s == falseSW = Nothing
+                  | True         = Just $ "(not " ++ cvtSW skolemMap s ++ ")"
+
+                 pos s
+                  | s == trueSW  = Nothing
+                  | s == falseSW = Just $ cvtSW skolemMap falseSW
+                  | True         = Just $ cvtSW skolemMap s
+
         skolemMap = M.fromList [(s, ss) | Right (s, ss) <- skolemInps, not (null ss)]
         tableMap  = IM.fromList $ map mkConstTable constTables ++ map mkSkTable skolemTables
           where mkConstTable (((t, _, _), _), _) = (t, "table" ++ show t)
                 mkSkTable    (((t, _, _), _), _) = (t, "table" ++ show t ++ forallArgs)
         asgns = F.toList asgnsSeq
-        mkLet (s, SBVApp (Label m) [e]) = "(let ((" ++ show s ++ " " ++ cvtSW     skolemMap          e ++ ")) ; " ++ m
-        mkLet (s, e)                    = "(let ((" ++ show s ++ " " ++ cvtExp rm skolemMap tableMap e ++ "))"
-        declConst useDefFun (s, c)
-          | useDefFun = ["(define-fun "   ++ varT ++ " " ++ cvtCW rm c ++ ")"]
-          | True      = [ "(declare-fun " ++ varT ++ ")"
-                        , "(assert (= "   ++ show s ++ " " ++ cvtCW rm c ++ "))"
-                        ]
-          where varT = show s ++ " " ++ swFunType [] s
+
+        mkAssign a
+          | null foralls = declDef cfg skolemMap tableMap a
+          | True         = letShift (mkLet a)
+
+        mkLet (s, SBVApp (Label m) [e]) = "(let ((" ++ show s ++ " " ++ cvtSW                skolemMap          e ++ ")) ; " ++ m
+        mkLet (s, e)                    = "(let ((" ++ show s ++ " " ++ cvtExp solverCaps rm skolemMap tableMap e ++ "))"
+
         userName s = case s `lookup` map snd inputs of
                         Just u  | show s /= u -> " ; tracks user variable " ++ show u
                         _ -> ""
-        -- following sorts are built-in; do not translate them:
-        builtInSort = (`elem` ["RoundingMode"])
-        declSort (s, _)
-          | builtInSort s           = []
-        declSort (s, Left  r ) = ["(declare-sort " ++ s ++ " 0)  ; N.B. Uninterpreted: " ++ r]
-        declSort (s, Right fs) = [ "(declare-datatypes () ((" ++ s ++ " " ++ unwords (map (\c -> "(" ++ c ++ ")") fs) ++ ")))"
-                                 , "(define-fun " ++ s ++ "_constrIndex ((x " ++ s ++ ")) Int"
-                                 ] ++ ["   " ++ body fs (0::Int)] ++ [")"]
-                where body []     _ = ""
-                      body [_]    i = show i
-                      body (c:cs) i = "(ite (= x " ++ c ++ ") " ++ show i ++ " " ++ body cs (i+1) ++ ")"
+
+-- | Declare new sorts
+declSort :: (String, Either String [String]) -> [String]
+declSort (s, _)
+  | s == "RoundingMode" -- built-in-sort; so don't declare.
+  = []
+declSort (s, Left  r ) = ["(declare-sort " ++ s ++ " 0)  ; N.B. Uninterpreted: " ++ r]
+declSort (s, Right fs) = [ "(declare-datatypes () ((" ++ s ++ " " ++ unwords (map (\c -> "(" ++ c ++ ")") fs) ++ ")))"
+                         , "(define-fun " ++ s ++ "_constrIndex ((x " ++ s ++ ")) Int"
+                         ] ++ ["   " ++ body fs (0::Int)] ++ [")"]
+        where body []     _ = ""
+              body [_]    i = show i
+              body (c:cs) i = "(ite (= x " ++ c ++ ") " ++ show i ++ " " ++ body cs (i+1) ++ ")"
+
+-- | Convert in a query context
+cvtInc :: SMTLibIncConverter [String]
+cvtInc inps ks consts arrs tbls (SBVPgm asgnsSeq) cfg =
+            -- sorts
+               concatMap declSort [(s, dt) | KUserSort s dt <- Set.toList ks]
+            -- constants
+            ++ map (declConst cfg) consts
+            -- inputs
+            ++ map declInp inps
+            -- arrays
+            ++ concat arrayConstants
+            -- tables
+            ++ concatMap constTable allTables
+            -- expressions
+            ++ map  (declDef cfg skolemMap tableMap) (F.toList asgnsSeq)
+            -- delayed equalities
+            ++ concatMap asrt arrayDelayeds
+  where -- NB. The below setting of skolemMap to empty is OK, since we do
+        -- not support queries in the context of skolemized variables
+        skolemMap = M.empty
+
+        rm = roundingMode cfg
+
+        declInp (s, _) = "(declare-fun " ++ show s ++ " () " ++ swType s ++ ")"
+
+        (arrayConstants, arrayDelayeds) = unzip $ map (declArray False (map fst consts) skolemMap) arrs
+
+        allTables = [(t, either id id (genTableData rm skolemMap (False, []) (map fst consts) t)) | t <- tbls]
+        tableMap  = IM.fromList $ map mkTable allTables
+          where mkTable (((t, _, _), _), _) = (t, "table" ++ show t)
+
+        asrt = map (\s -> "(assert " ++ s ++ ")")
+
+declDef :: SMTConfig -> SkolemMap -> TableMap -> (SW, SBVExpr) -> String
+declDef cfg skolemMap tableMap (s, expr) =
+        case expr of
+          SBVApp  (Label m) [e] -> defineFun (s, cvtSW          skolemMap          e) (Just m)
+          e                     -> defineFun (s, cvtExp caps rm skolemMap tableMap e) Nothing
+  where caps = capabilities (solver cfg)
+        rm   = roundingMode cfg
+
+defineFun :: (SW, String) -> Maybe String -> String
+defineFun (s, def) mbComment = "(define-fun "   ++ varT ++ " " ++ def ++ ")" ++ cmnt
+  where varT      = show s ++ " " ++ swFunType [] s
+        cmnt      = maybe "" (" ; " ++) mbComment
+
+declConst :: SMTConfig -> (SW, CW) -> String
+declConst cfg (s, c) = defineFun (s, cvtCW (roundingMode cfg) c) Nothing
 
 declUI :: (String, SBVType) -> [String]
 declUI (i, t) = ["(declare-fun " ++ i ++ " " ++ cvtType t ++ ")"]
@@ -245,16 +326,14 @@ genTableData rm skolemMap (_quantified, args) consts ((i, aknd, _), elts)
 
 -- TODO: We currently do not support non-constant arrays when quantifiers are present, as
 -- we might have to skolemize those. Implement this properly.
--- The difficulty is with the ArrayReset/Mutate/Merge: We have to postpone an init if
+-- The difficulty is with the Mutate/Merge: We have to postpone an init if
 -- the components are themselves postponed, so this cannot be implemented as a simple map.
 declArray :: Bool -> [SW] -> SkolemMap -> (Int, ArrayInfo) -> ([String], [String])
-declArray quantified consts skolemMap (i, (_, (aKnd, bKnd), ctx)) = (adecl : map wrap pre, map snd post)
+declArray quantified consts skolemMap (i, (_, (aKnd, bKnd), ctx)) = (adecl : map (wrap . snd) pre, map snd post)
   where topLevel = not quantified || case ctx of
-                                       ArrayFree Nothing -> True
-                                       ArrayFree (Just sw) -> sw `elem` consts
-                                       ArrayReset _ sw     -> sw `elem` consts
-                                       ArrayMutate _ a b   -> all (`elem` consts) [a, b]
-                                       ArrayMerge c _ _    -> c `elem` consts
+                                       ArrayFree         -> True
+                                       ArrayMutate _ a b -> all (`elem` consts) [a, b]
+                                       ArrayMerge c _ _  -> c `elem` consts
         (pre, post) = partition fst ctxInfo
         nm = "array_" ++ show i
         ssw sw
@@ -264,17 +343,10 @@ declArray quantified consts skolemMap (i, (_, (aKnd, bKnd), ctx)) = (adecl : map
          = tbd "Non-constant array initializer in a quantified context"
         adecl = "(declare-fun " ++ nm ++ " () (Array " ++ smtType aKnd ++ " " ++ smtType bKnd ++ "))"
         ctxInfo = case ctx of
-                    ArrayFree Nothing   -> []
-                    ArrayFree (Just sw) -> declA sw
-                    ArrayReset _ sw     -> declA sw
+                    ArrayFree         -> []
                     ArrayMutate j a b -> [(all (`elem` consts) [a, b], "(= " ++ nm ++ " (store array_" ++ show j ++ " " ++ ssw a ++ " " ++ ssw b ++ "))")]
                     ArrayMerge  t j k -> [(t `elem` consts,            "(= " ++ nm ++ " (ite " ++ ssw t ++ " array_" ++ show j ++ " array_" ++ show k ++ "))")]
-        declA sw = let iv = nm ++ "_freeInitializer"
-                   in [ (True,             "(declare-fun " ++ iv ++ " () " ++ smtType aKnd ++ ")")
-                      , (sw `elem` consts, "(= (select " ++ nm ++ " " ++ iv ++ ") " ++ ssw sw ++ ")")
-                      ]
-        wrap (False, s) = s
-        wrap (True, s)  = "(assert " ++ s ++ ")"
+        wrap s = "(assert " ++ s ++ ")"
 
 swType :: SW -> String
 swType s = smtType (kindOf s)
@@ -312,29 +384,43 @@ cvtCW = cwToSMTLib
 getTable :: TableMap -> Int -> String
 getTable m i
   | Just tn <- i `IM.lookup` m = tn
-  | True                       = error $ "SBV.SMTLib2: Cannot locate table " ++ show i
+  | True                       = "table" ++ show i  -- constant tables are always named this way
 
-cvtExp :: RoundingMode -> SkolemMap -> TableMap -> SBVExpr -> String
-cvtExp rm skolemMap tableMap expr@(SBVApp _ arguments) = sh expr
+cvtExp :: SolverCapabilities -> RoundingMode -> SkolemMap -> TableMap -> SBVExpr -> String
+cvtExp caps rm skolemMap tableMap expr@(SBVApp _ arguments) = sh expr
   where ssw = cvtSW skolemMap
-        bvOp     = all isBounded       arguments
-        intOp    = any isInteger       arguments
-        realOp   = any isReal          arguments
-        doubleOp = any isDouble        arguments
-        floatOp  = any isFloat         arguments
-        boolOp   = all isBoolean       arguments
+
+        supportsPB = supportsPseudoBooleans caps
+
+        bvOp     = all isBounded arguments
+        intOp    = any isInteger arguments
+        realOp   = any isReal    arguments
+        doubleOp = any isDouble  arguments
+        floatOp  = any isFloat   arguments
+        boolOp   = all isBoolean arguments
+
         bad | intOp = error $ "SBV.SMTLib2: Unsupported operation on unbounded integers: " ++ show expr
             | True  = error $ "SBV.SMTLib2: Unsupported operation on real values: " ++ show expr
+
         ensureBVOrBool = bvOp || boolOp || bad
         ensureBV       = bvOp || bad
+
         addRM s = s ++ " " ++ smtRoundingMode rm
+
+        -- lift a binary op
         lift2  o _ [x, y] = "(" ++ o ++ " " ++ x ++ " " ++ y ++ ")"
         lift2  o _ sbvs   = error $ "SBV.SMTLib2.sh.lift2: Unexpected arguments: "   ++ show (o, sbvs)
+
+        -- lift an arbitrary arity operator
+        liftN o _ xs = "(" ++ o ++ " " ++ unwords xs ++ ")"
+
         -- lift a binary operation with rounding-mode added; used for floating-point arithmetic
         lift2WM o fo | doubleOp || floatOp = lift2 (addRM fo)
                      | True                = lift2 o
+
         lift1FP o fo | doubleOp || floatOp = lift1 fo
                      | True                = lift1 o
+
         liftAbs sgned args | doubleOp || floatOp = lift1 "fp.abs" sgned args
                            | intOp               = lift1 "abs"    sgned args
                            | bvOp, sgned         = mkAbs (head args) "bvslt" "bvneg"
@@ -344,32 +430,51 @@ cvtExp rm skolemMap tableMap expr@(SBVApp _ arguments) = sh expr
                   where ltz = "(" ++ cmp ++ " " ++ x ++ " " ++ z ++ ")"
                         nx  = "(" ++ neg ++ " " ++ x ++ ")"
                         z   = cvtCW rm (mkConstCW (kindOf (head arguments)) (0::Integer))
+
         lift2B bOp vOp
           | boolOp = lift2 bOp
           | True   = lift2 vOp
+
         lift1B bOp vOp
           | boolOp = lift1 bOp
           | True   = lift1 vOp
+
         eqBV  = lift2 "="
-        neqBV = lift2 "distinct"
+        neqBV = liftN "distinct"
+
         equal sgn sbvs
           | doubleOp = lift2 "fp.eq" sgn sbvs
           | floatOp  = lift2 "fp.eq" sgn sbvs
-          | True     = lift2 "=" sgn sbvs
+          | True     = lift2 "="     sgn sbvs
+
         notEqual sgn sbvs
-          | doubleOp = "(not " ++ equal sgn sbvs ++ ")"
-          | floatOp  = "(not " ++ equal sgn sbvs ++ ")"
-          | True     = lift2 "distinct" sgn sbvs
+          | doubleOp = liftP sbvs
+          | floatOp  = liftP sbvs
+          | True     = liftN "distinct" sgn sbvs
+          where liftP [_, _] = "(not " ++ equal sgn sbvs ++ ")"
+                liftP args   = "(and " ++ unwords (walk args) ++ ")"
+
+                walk []     = []
+                walk (e:es) = map (pair e) es ++ walk es
+
+                pair e1 e2  = "(not (fp.eq " ++ e1 ++ " " ++ e2 ++ "))"
+
         lift2S oU oS sgn = lift2 (if sgn then oS else oU) sgn
+        liftNS oU oS sgn = liftN (if sgn then oS else oU) sgn
+
         lift2Cmp o fo | doubleOp || floatOp = lift2 fo
                       | True                = lift2 o
+
         unintComp o [a, b]
           | KUserSort s (Right _) <- kindOf (head arguments)
-          = let idx v = "(" ++ s ++ "_constrIndex " ++ " " ++ v ++ ")" in "(" ++ o ++ " " ++ idx a ++ " " ++ idx b ++ ")"
+          = let idx v = "(" ++ s ++ "_constrIndex " ++ v ++ ")" in "(" ++ o ++ " " ++ idx a ++ " " ++ idx b ++ ")"
         unintComp o sbvs = error $ "SBV.SMT.SMTLib2.sh.unintComp: Unexpected arguments: "   ++ show (o, sbvs)
+
         lift1  o _ [x]    = "(" ++ o ++ " " ++ x ++ ")"
         lift1  o _ sbvs   = error $ "SBV.SMT.SMTLib2.sh.lift1: Unexpected arguments: "   ++ show (o, sbvs)
+
         sh (SBVApp Ite [a, b, c]) = "(ite " ++ ssw a ++ " " ++ ssw b ++ " " ++ ssw c ++ ")"
+
         sh (SBVApp (LkUp (t, aKnd, _, l) i e) [])
           | needsCheck = "(ite " ++ cond ++ ssw e ++ " " ++ lkUp ++ ")"
           | True       = lkUp
@@ -396,30 +501,33 @@ cvtExp rm skolemMap tableMap expr@(SBVApp _ arguments) = sh expr
                 mkCnst = cvtCW rm . mkConstCW (kindOf i)
                 le0  = "(" ++ less ++ " " ++ ssw i ++ " " ++ mkCnst 0 ++ ")"
                 gtl  = "(" ++ leq  ++ " " ++ mkCnst l ++ " " ++ ssw i ++ ")"
+
         sh (SBVApp (KindCast f t) [a]) = handleKindCast f t (ssw a)
+
         sh (SBVApp (ArrEq i j) [])  = "(= array_" ++ show i ++ " array_" ++ show j ++")"
         sh (SBVApp (ArrRead i) [a]) = "(select array_" ++ show i ++ " " ++ ssw a ++ ")"
+
         sh (SBVApp (Uninterpreted nm) [])   = nm
         sh (SBVApp (Uninterpreted nm) args) = "(" ++ nm ++ " " ++ unwords (map ssw args) ++ ")"
+
         sh (SBVApp (Extract i j) [a]) | ensureBV = "((_ extract " ++ show i ++ " " ++ show j ++ ") " ++ ssw a ++ ")"
+
         sh (SBVApp (Rol i) [a])
-           | bvOp  = rot  ssw "rotate_left"  i a
-           | intOp = sh (SBVApp (Shl i) [a])       -- Haskell treats rotateL as shiftL for unbounded values
+           | bvOp  = rot ssw "rotate_left"  i a
            | True  = bad
+
         sh (SBVApp (Ror i) [a])
            | bvOp  = rot  ssw "rotate_right" i a
-           | intOp = sh (SBVApp (Shr i) [a])     -- Haskell treats rotateR as shiftR for unbounded values
            | True  = bad
-        sh (SBVApp (Shl i) [a])
-           | bvOp   = shft rm ssw "bvshl"  "bvshl"  i a
-           | i < 0  = sh (SBVApp (Shr (-i)) [a])  -- flip sign/direction
-           | intOp  = "(* " ++ ssw a ++ " " ++ show (bit i :: Integer) ++ ")"  -- Implement shiftL by multiplication by 2^i
+
+        sh (SBVApp Shl [a, i])
+           | bvOp   = shft ssw "bvshl"  "bvshl" a i
            | True   = bad
-        sh (SBVApp (Shr i) [a])
-           | bvOp  = shft rm ssw "bvlshr" "bvashr" i a
-           | i < 0 = sh (SBVApp (Shl (-i)) [a])  -- flip sign/direction
-           | intOp = "(div " ++ ssw a ++ " " ++ show (bit i :: Integer) ++ ")"  -- Implement shiftR by division by 2^i
+
+        sh (SBVApp Shr [a, i])
+           | bvOp  = shft ssw "bvlshr" "bvashr" a i
            | True  = bad
+
         sh (SBVApp op args)
           | Just f <- lookup op smtBVOpTable, ensureBVOrBool
           = f (any hasSign args) (map ssw args)
@@ -432,9 +540,17 @@ cvtExp rm skolemMap tableMap expr@(SBVApp _ arguments) = sh expr
                                , (Not,  lift1B "not" "bvnot")
                                , (Join, lift2 "concat")
                                ]
-        sh (SBVApp (Label _)                       [a]) = cvtSW skolemMap a  -- This won't be reached; but just in case!
+
+        sh (SBVApp (Label m) [a]) = curry named (Just m) $ cvtSW skolemMap a  -- This won't be reached; but just in case!
+
         sh (SBVApp (IEEEFP (FP_Cast kFrom kTo m)) args) = handleFPCast kFrom kTo (ssw m) (unwords (map ssw args))
         sh (SBVApp (IEEEFP w                    ) args) = "(" ++ show w ++ " " ++ unwords (map ssw args) ++ ")"
+
+        sh (SBVApp (PseudoBoolean pb) args)
+          | supportsPB = handlePB pb args'
+          | True       = reducePB pb args'
+          where args' = map ssw args
+
         sh inp@(SBVApp op args)
           | intOp, Just f <- lookup op smtOpIntTable
           = f True (map ssw args)
@@ -499,7 +615,7 @@ cvtExp rm skolemMap tableMap expr@(SBVApp _ arguments) = sh expr
                                     ]
                 -- equality and comparisons are the only thing that works on uninterpreted sorts
                 uninterpretedTable = [ (Equal,       lift2S "="        "="        True)
-                                     , (NotEqual,    lift2S "distinct" "distinct" True)
+                                     , (NotEqual,    liftNS "distinct" "distinct" True)
                                      , (LessThan,    unintComp "<")
                                      , (GreaterThan, unintComp ">")
                                      , (LessEq,      unintComp "<=")
@@ -576,11 +692,9 @@ handleFPCast kFrom kTo rm input
 rot :: (SW -> String) -> String -> Int -> SW -> String
 rot ssw o c x = "((_ " ++ o ++ " " ++ show c ++ ") " ++ ssw x ++ ")"
 
-shft :: RoundingMode -> (SW -> String) -> String -> String -> Int -> SW -> String
-shft rm ssw oW oS c x = "(" ++ o ++ " " ++ ssw x ++ " " ++ cvtCW rm c' ++ ")"
-   where s  = hasSign x
-         c' = mkConstCW (kindOf x) c
-         o  = if s then oS else oW
+shft :: (SW -> String) -> String -> String -> SW -> SW -> String
+shft ssw oW oS x c = "(" ++ o ++ " " ++ ssw x ++ " " ++ ssw c ++ ")"
+   where o = if hasSign x then oS else oW
 
 -- Various casts
 handleKindCast :: Kind -> Kind -> String -> String
@@ -638,3 +752,30 @@ handleKindCast kFrom kTo a
         signExtend i = "((_ sign_extend " ++ show i ++  ") "  ++ a ++ ")"
         zeroExtend i = "((_ zero_extend " ++ show i ++  ") "  ++ a ++ ")"
         extract    i = "((_ extract "     ++ show i ++ " 0) " ++ a ++ ")"
+
+-- Translation of pseudo-booleans, in case the solver supports them
+handlePB :: PBOp -> [String] -> String
+handlePB (PB_AtMost  k) args = "((_ at-most "  ++ show k                                             ++ ") " ++ unwords args ++ ")"
+handlePB (PB_AtLeast k) args = "((_ at-least " ++ show k                                             ++ ") " ++ unwords args ++ ")"
+handlePB (PB_Exactly k) args = "((_ pbeq "     ++ unwords (map show (k : replicate (length args) 1)) ++ ") " ++ unwords args ++ ")"
+handlePB (PB_Eq cs   k) args = "((_ pbeq "     ++ unwords (map show (k : cs))                        ++ ") " ++ unwords args ++ ")"
+handlePB (PB_Le cs   k) args = "((_ pble "     ++ unwords (map show (k : cs))                        ++ ") " ++ unwords args ++ ")"
+handlePB (PB_Ge cs   k) args = "((_ pbge "     ++ unwords (map show (k : cs))                        ++ ") " ++ unwords args ++ ")"
+
+-- Translation of pseudo-booleans, in case the solver does *not* support them
+reducePB :: PBOp -> [String] -> String
+reducePB op args = case op of
+                     PB_AtMost  k -> "(<= " ++ addIf (repeat 1) ++ " " ++ show k ++ ")"
+                     PB_AtLeast k -> "(>= " ++ addIf (repeat 1) ++ " " ++ show k ++ ")"
+                     PB_Exactly k -> "(=  " ++ addIf (repeat 1) ++ " " ++ show k ++ ")"
+                     PB_Le cs   k -> "(<= " ++ addIf cs         ++ " " ++ show k ++ ")"
+                     PB_Ge cs   k -> "(>= " ++ addIf cs         ++ " " ++ show k ++ ")"
+                     PB_Eq cs   k -> "(=  " ++ addIf cs         ++ " " ++ show k ++ ")"
+
+  where addIf :: [Int] -> String
+        addIf cs = "(+ " ++ unwords ["(ite " ++ a ++ " " ++ show c ++ " 0)" | (a, c) <- zip args cs] ++ ")"
+
+-- Add a named annotation
+named :: (Maybe String, String) -> String
+named (Nothing, x) = x
+named (Just nm, x) = annotateWithName nm x
