@@ -1,6 +1,6 @@
 -----------------------------------------------------------------------------
 -- |
--- Module      :  Data.SBV.BitVectors.Polynomials
+-- Module      :  Data.SBV.Core.Polynomials
 -- Copyright   :  (c) Levent Erkok
 -- License     :  BSD3
 -- Maintainer  :  erkokl@gmail.com
@@ -14,16 +14,19 @@
 {-# LANGUAGE PatternGuards        #-}
 {-# LANGUAGE TypeSynonymInstances #-}
 
-module Data.SBV.Tools.Polynomial (Polynomial(..), crc, crcBV, ites, mdp, addPoly) where
+module Data.SBV.Tools.Polynomial (
+        -- * Polynomial arithmetic and CRCs
+        Polynomial(..), crc, crcBV, ites, mdp, addPoly
+        ) where
 
 import Data.Bits  (Bits(..))
 import Data.List  (genericTake)
 import Data.Maybe (fromJust, fromMaybe)
 import Data.Word  (Word8, Word16, Word32, Word64)
 
-import Data.SBV.BitVectors.Data
-import Data.SBV.BitVectors.Model
-import Data.SBV.BitVectors.Splittable
+import Data.SBV.Core.Data
+import Data.SBV.Core.Model
+
 import Data.SBV.Utils.Boolean
 
 -- | Implements polynomial addition, multiplication, division, and modulus operations
@@ -130,7 +133,7 @@ ites s xs ys
 
 -- | Multiply two polynomials and reduce by the third (concrete) irreducible, given by its coefficients.
 -- See the remarks for the 'pMult' function for this design choice
-polyMult :: (Num a, Bits a, SymWord a, FromBits (SBV a)) => (SBV a, SBV a, [Int]) -> SBV a
+polyMult :: SFiniteBits a => (SBV a, SBV a, [Int]) -> SBV a
 polyMult (x, y, red)
   | isReal x
   = error $ "SBV.polyMult: Received a real value: " ++ show x
@@ -145,7 +148,7 @@ polyMult (x, y, red)
         mul _  []     ps = ps
         mul as (b:bs) ps = mul (false:as) bs (ites b (as `addPoly` ps) ps)
 
-polyDivMod :: (Num a, Bits a, SymWord a, FromBits (SBV a)) => SBV a -> SBV a -> (SBV a, SBV a)
+polyDivMod :: SFiniteBits a => SBV a -> SBV a -> (SBV a, SBV a)
 polyDivMod x y
    | isReal x
    = error $ "SBV.polyDivMod: Received a real value: " ++ show x
@@ -236,7 +239,7 @@ crcBV n m p = take n $ go (replicate n false) (m ++ replicate n false)
 
 -- | Compute CRC's over polynomials, i.e., symbolic words. The first
 -- 'Int' argument plays the same role as the one in the 'crcBV' function.
-crc :: (FromBits (SBV a), FromBits (SBV b), Num a, Num b, Bits a, Bits b, SymWord a, SymWord b) => Int -> SBV a -> SBV b -> SBV b
+crc :: (SFiniteBits a, SFiniteBits b) => Int -> SBV a -> SBV b -> SBV b
 crc n m p
   | isReal m || isReal p
   = error $ "SBV.crc: Received a real value: " ++ show (m, p)
